"""Hydra types module."""

from typing import NewType

HydraAccessToken = NewType("HydraAccessToken", str)
HydraClientId = NewType("HydraClientId", str)
HydraClientSecret = NewType("HydraClientSecret", str)

__all__: list[str] = [
    "HydraAccessToken",
    "HydraClientId",
    "HydraClientSecret",
]



"""Provides enums for Kratos."""

from enum import StrEnum


class AuthenticatorAssuranceLevelEnum(StrEnum):
    """Enum for Authenticator Assurance Level (AAL)."""

    AAL0 = "aal0"
    AAL1 = "aal1"
    AAL2 = "aal2"
    AAL3 = "aal3"


class AuthenticationMethodEnum(StrEnum):
    """Enum for Authentication Method."""

    PASSWORD = "password"
    OIDC = "oidc"
    TOTP = "totp"
    LOOKUP_SECRET = "lookup_secret"
    WEBAUTHN = "webauthn"
    CODE = "code"
    PASSKEY = "passkey"
    PROFILE = "profile"
    SAML = "saml"
    LINK_RECOVERY = "link_recovery"
    CODE_RECOVERY = "code_recovery"


class KratosIdentityStateEnum(StrEnum):
    """Enum for Identity State."""

    ACTIVE = "active"
    INACTIVE = "inactive"


class KratosFlowTypeEnum(StrEnum):
    """Enum for Flow Type."""

    API = "api"
    BROWSER = "browser"


class KratosIdentityPatchOpEnum(StrEnum):
    """Enum for Kratos identity patch operation."""

    ADD = "add"
    REMOVE = "remove"
    REPLACE = "replace"
    MOVE = "move"
    COPY = "copy"
    TEST = "test"



"""Aiohttp client factory."""

import asyncio
import logging
import os
import ssl
import time
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from importlib.util import find_spec
from typing import Any

import aiohttp
import certifi
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.trace import TracerProvider
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins.aiohttp.configs import HttpServiceDependencyConfig
from fastapi_factory_utilities.core.plugins.aiohttp.exceptions import AioHttpClientError

_logger: BoundLogger = get_logger()


class AioHttpClientResource:
    """Aiohttp client resource.

    Objectives:
    - Provide a resource to acquire and release Aiohttp client sessions.
    - Build the TCP connector with the dependency configuration.
    - Instrument the client session with OpenTelemetry.
    - Provide a graceful shutdown of the client sessions and the TCP connector.
    """

    @classmethod
    def build_ssl_context(cls, dependency_config: HttpServiceDependencyConfig) -> ssl.SSLContext | bool:
        """Build the SSL context.

        Args:
            dependency_config (HttpServiceDependencyConfig): The dependency configuration.

        Returns:
            ssl.SSLContext | bool: The SSL context.
        """
        ssl_context: ssl.SSLContext | bool = False

        # If the SSL verification is disabled, return False to disable the SSL verification.
        if dependency_config.verify_ssl is False:
            return ssl_context

        # If the SSL CA path is provided, use it to create the SSL context.
        if dependency_config.ssl_ca_path is not None:
            if not os.path.exists(dependency_config.ssl_ca_path):
                raise AioHttpClientError(f"SSL CA path {dependency_config.ssl_ca_path} does not exist.")
            ssl_context = ssl.create_default_context(cafile=dependency_config.ssl_ca_path)
        else:
            # If no SSL CA path is provided, use the default SSL CA path from certifi.
            ssl_context = ssl.create_default_context(cafile=certifi.where())

        if dependency_config.ssl_certfile is None or dependency_config.ssl_keyfile is None:
            return ssl_context

        if not os.path.exists(dependency_config.ssl_certfile):
            raise AioHttpClientError(f"SSL certificate file {dependency_config.ssl_certfile} does not exist.")
        if not os.path.exists(dependency_config.ssl_keyfile):
            raise AioHttpClientError(f"SSL key file {dependency_config.ssl_keyfile} does not exist.")

        # Load the SSL certificate and key files into the SSL context.
        if dependency_config.ssl_keyfile_password is not None:
            ssl_context.load_cert_chain(
                dependency_config.ssl_certfile,
                dependency_config.ssl_keyfile,
                password=dependency_config.ssl_keyfile_password,
            )
        else:
            ssl_context.load_cert_chain(dependency_config.ssl_certfile, dependency_config.ssl_keyfile)

        return ssl_context

    @classmethod
    def build_tcp_connector(cls, dependency_config: HttpServiceDependencyConfig) -> aiohttp.TCPConnector:
        """Build the TCP connector.

        Args:
            dependency_config (HttpServiceDependencyConfig): The dependency configuration.

        Returns:
            aiohttp.TCPConnector: The TCP connector.
        """
        ssl_context: ssl.SSLContext | bool = cls.build_ssl_context(dependency_config=dependency_config)
        return aiohttp.TCPConnector(
            limit=dependency_config.limit,
            limit_per_host=dependency_config.limit_per_host,
            use_dns_cache=dependency_config.use_dns_cache,
            ttl_dns_cache=dependency_config.ttl_dns_cache,
            ssl=ssl_context,
        )

    @classmethod
    def build_trace_config(
        cls, tracer_provider: TracerProvider | None, meter_provider: MeterProvider | None
    ) -> aiohttp.TraceConfig | None:
        """Build the trace config.

        Args:
            tracer_provider: The tracer provider. If None, OpenTelemetry tracing is disabled.
            meter_provider: The meter provider. If None, OpenTelemetry metrics are disabled.

        Returns:
            aiohttp.TraceConfig: The trace config, or None if providers are not available.
        """
        if tracer_provider is None or meter_provider is None:
            return None
        if find_spec(name="opentelemetry.instrumentation.aiohttp_client"):
            from opentelemetry.instrumentation.aiohttp_client import (  # pylint: disable=import-outside-toplevel # noqa: PLC0415 # pyright: ignore
                create_trace_config,
            )

            return create_trace_config(tracer_provider=tracer_provider, meter_provider=meter_provider)
        return None

    def __init__(self, dependency_config: HttpServiceDependencyConfig) -> None:
        """Initialize the Aiohttp client resource.

        Args:
            dependency_config (HttpServiceDependencyConfig): The dependency configuration.

        Returns:
            None
        """
        self._dependency_config: HttpServiceDependencyConfig = dependency_config
        self._tcp_connector: aiohttp.TCPConnector | None = None
        self._client_sessions: list[aiohttp.ClientSession] = []
        self._tracer_provider: TracerProvider | None = None
        self._meter_provider: MeterProvider | None = None
        self._must_be_closed: bool = False

    def on_load(self) -> None:
        """On load.

        This method is called when the resource is loaded.
        Currently, no initialization is required at load time.
        """

    async def on_startup(
        self, tracer_provider: TracerProvider | None = None, meter_provider: MeterProvider | None = None
    ) -> None:
        """On startup.

        Args:
            tracer_provider: The tracer provider. If None, OpenTelemetry tracing is disabled.
            meter_provider: The meter provider. If None, OpenTelemetry metrics are disabled.

        Returns:
            None
        """
        self._tracer_provider = tracer_provider
        self._meter_provider = meter_provider
        if self._tcp_connector is None:
            self._tcp_connector = self.build_tcp_connector(dependency_config=self._dependency_config)

    async def on_shutdown(self) -> None:
        """On shutdown.

        Args:
            None

        Returns:
            None
        """
        # Wait for all client sessions to be released.
        initial_time = time.time()
        self._must_be_closed = True
        # Wait for all client sessions to be released or the timeout is reached.
        while self._client_sessions and time.time() - initial_time < self._dependency_config.graceful_shutdown_timeout:
            await asyncio.sleep(0.5)
        # Close all client sessions remaining.
        for session in self._client_sessions:
            try:
                await session.close()
            except (OSError, aiohttp.ClientError) as exception:
                _logger.log(
                    level=logging.WARNING,
                    event="Failed to close client session during shutdown",
                    exception=str(exception),
                )
        # Close the TCP connector.
        if self._tcp_connector is not None:
            try:
                await self._tcp_connector.close()
            except (OSError, aiohttp.ClientError) as exception:
                _logger.log(
                    level=logging.WARNING,
                    event="Failed to close TCP connector during shutdown",
                    exception=str(exception),
                )

    @asynccontextmanager
    async def acquire_client_session(self, **kwargs: Any) -> AsyncGenerator[aiohttp.ClientSession, None]:
        """Acquire the Aiohttp client session.

        Args:
            **kwargs: Any additional keyword arguments to pass to the ClientSession constructor.

        Returns:
            AsyncGenerator[aiohttp.ClientSession, None]: The Aiohttp client session.

        Raises:
            RuntimeError: If the TCP connector is not initialized.
            ValueError: If the connector is already provided in kwargs.
        """
        if self._tcp_connector is None:
            raise AioHttpClientError("TCP connector is not initialized. Call on_startup() first.")
        if self._tcp_connector.closed and self._must_be_closed:
            raise AioHttpClientError("TCP connector is closed.")
        if self._tcp_connector.closed and not self._must_be_closed:
            _logger.warning(
                "TCP connector has been closed but the resource is not being shutdown. Rebuilding the TCP connector.",
            )
            self._tcp_connector = self.build_tcp_connector(dependency_config=self._dependency_config)
        if "connector" in kwargs:
            raise ValueError("The connector is already provided.")

        kwargs["connector"] = self._tcp_connector
        # Set the connector owner to False to avoid closing the connector when the session is closed.
        kwargs["connector_owner"] = False

        if self._dependency_config.url is not None:
            kwargs["base_url"] = str(self._dependency_config.url)

        trace_config: aiohttp.TraceConfig | None = self.build_trace_config(
            tracer_provider=self._tracer_provider, meter_provider=self._meter_provider
        )

        if trace_config is not None:
            kwargs["trace_configs"] = [trace_config]

        async with aiohttp.ClientSession(**kwargs) as session:
            self._client_sessions.append(session)
            try:
                yield session
            finally:
                self._client_sessions.remove(session)



"""Provides the types for the Aiopika plugin."""

from typing import NewType

RoutingKey = NewType("RoutingKey", str)
ExchangeName = NewType("ExchangeName", str)
QueueName = NewType("QueueName", str)



"""Aiohttp plugin."""

from importlib.util import find_spec

from .configs import HttpServiceDependencyConfig
from .depends import AioHttpResourceDepends
from .exceptions import AioHttpClientError, AioHttpClientResourceNotFoundError, UnableToReadHttpDependencyConfigError
from .plugins import AioHttpClientPlugin
from .resources import AioHttpClientResource

__all__: list[str] = []  # pylint: disable=invalid-name

# Add mockers helpers only if pytest is installed
if find_spec(name="pytest"):
    from .mockers import build_mocked_aiohttp_resource, build_mocked_aiohttp_response

    __all__ += [
        "build_mocked_aiohttp_resource",
        "build_mocked_aiohttp_response",
    ]

__all__ += [
    "AioHttpClientError",
    "AioHttpClientPlugin",
    "AioHttpClientResource",
    "AioHttpClientResourceNotFoundError",
    "AioHttpResourceDepends",
    "HttpServiceDependencyConfig",
    "UnableToReadHttpDependencyConfigError",
]



"""Provides the JWT bearer token validator."""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, cast

from fastapi_factory_utilities.core.security.types import JWTToken
from fastapi_factory_utilities.core.services.hydra import (
    HydraAccessToken,
    HydraIntrospectGenericService,
    HydraOperationError,
    HydraTokenIntrospectObject,
)

from .exceptions import InvalidJWTError
from .objects import JWTPayload

JWTBearerPayloadGeneric = TypeVar("JWTBearerPayloadGeneric", bound=JWTPayload)
HydraIntrospectObjectGeneric = TypeVar("HydraIntrospectObjectGeneric", bound=HydraTokenIntrospectObject)


class JWTVerifierAbstract(ABC, Generic[JWTBearerPayloadGeneric]):
    """JWT verifier."""

    @abstractmethod
    async def verify(
        self,
        jwt_token: JWTToken,
        jwt_payload: JWTBearerPayloadGeneric,
    ) -> None:
        """Verify the JWT bearer token.

        Args:
            jwt_token (JWTToken): The JWT bearer token.
            jwt_payload (JWTBearerPayloadGeneric): The JWT bearer payload.

        Raises:
            NotVerifiedJWTError: If the JWT bearer token is not verified.
        """
        raise NotImplementedError()


class GenericHydraJWTVerifier(
    JWTVerifierAbstract[JWTBearerPayloadGeneric], Generic[JWTBearerPayloadGeneric, HydraIntrospectObjectGeneric]
):
    """Generic Hydra JWT verifier."""

    def __init__(self, hydra_introspect_service: HydraIntrospectGenericService[HydraIntrospectObjectGeneric]) -> None:
        """Initialize the Generic Hydra JWT verifier."""
        self._hydra_introspect_service: HydraIntrospectGenericService[HydraIntrospectObjectGeneric] = (
            hydra_introspect_service
        )
        self._introspect_object: HydraIntrospectObjectGeneric | None = None

    @property
    def introspect_object(self) -> HydraIntrospectObjectGeneric:
        """Get the introspect object.

        Returns:
            HydraTokenIntrospectObject: The introspect object.
        """
        assert self._introspect_object is not None
        return self._introspect_object

    async def verify(self, jwt_token: JWTToken, jwt_payload: JWTBearerPayloadGeneric) -> None:
        """Verify the JWT token.

        Args:
            jwt_token: The JWT token.
            jwt_payload: The JWT payload.
        """
        try:
            self._introspect_object = await self._hydra_introspect_service.introspect(
                token=cast(HydraAccessToken, jwt_token)
            )
        except HydraOperationError as e:
            raise InvalidJWTError("Failed to introspect the JWT token") from e

        if self._introspect_object.active is False:
            raise InvalidJWTError("JWT token is not active")


class JWTNoneVerifier(JWTVerifierAbstract[JWTPayload]):
    """JWT none verifier."""

    async def verify(self, jwt_token: JWTToken, jwt_payload: JWTPayload) -> None:
        """Verify the JWT bearer token.

        Args:
            jwt_token (JWTToken): The JWT bearer token.
            jwt_payload (JWTBearerPayload): The JWT bearer payload.

        Raises:
            NotVerifiedJWTError: If the JWT bearer token is not verified.
        """
        return



"""Provides the exceptions for the audit service."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class AuditServiceError(FastAPIFactoryUtilitiesError):
    """Audit service error."""



"""Provides the JWT bearer token decoders.

Can be implemented to support different JWT bearer token formats or additional claims.
https://www.iana.org/assignments/jwt/jwt.xhtml#claims
"""

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, get_args

from jwt import ExpiredSignatureError, InvalidTokenError, decode, get_unverified_header
from jwt.api_jwk import PyJWK
from pydantic import ValidationError

from fastapi_factory_utilities.core.security.types import JWTToken, OAuth2Issuer, OAuth2Subject

from .configs import JWTBearerAuthenticationConfig
from .exceptions import ExpiredJWTError, InvalidJWTError, InvalidJWTPayploadError
from .objects import JWTPayload
from .stores import JWKStoreAbstract

JWTBearerPayloadGeneric = TypeVar("JWTBearerPayloadGeneric", bound=JWTPayload)


async def decode_jwt_token_payload(
    jwt_token: JWTToken,
    public_key: PyJWK,
    jwt_bearer_authentication_config: JWTBearerAuthenticationConfig,
    subject: OAuth2Subject | None = None,
    issuer: OAuth2Issuer | None = None,
) -> dict[str, Any]:
    """Decode the JWT bearer token payload.

    Args:
        jwt_token (JWTToken): The JWT bearer token.
        public_key (PyJWK): The public key.
        jwt_bearer_authentication_config (JWTBearerAuthenticationConfig): The JWT bearer authentication configuration.
        subject (OAuth2Subject | None): The subject.
        issuer (OAuth2Issuer | None): The issuer.

    Returns:
        dict[str, Any]: The decoded JWT bearer token payload.

    Raises:
        ExpiredJWTError: If the JWT bearer token is expired.
        InvalidJWTError: If the JWT bearer token is invalid.
    """
    # Additional kwargs for the decode function
    kwargs: dict[str, Any] = {}
    kwargs["issuer"] = issuer or jwt_bearer_authentication_config.issuer
    if jwt_bearer_authentication_config.authorized_audiences:
        kwargs["audience"] = jwt_bearer_authentication_config.authorized_audiences
    if subject:
        kwargs["subject"] = subject
    # Decode the JWT bearer token payload
    try:
        return decode(
            jwt=jwt_token,
            key=public_key,
            algorithms=jwt_bearer_authentication_config.authorized_algorithms,
            options={"verify_signature": True},
            **kwargs,
        )
    except ExpiredSignatureError as e:
        raise ExpiredJWTError("The JWT bearer token is expired") from e
    except InvalidTokenError as e:
        raise InvalidJWTError("Failed to decode the JWT bearer token payload") from e


class JWTBearerTokenDecoderAbstract(ABC, Generic[JWTBearerPayloadGeneric]):
    """JWT bearer token decoder."""

    def get_kid_from_jwt_unsafe_header(self, jwt_token: JWTToken) -> str:
        """Get the kid from the JWT header.

        Args:
            jwt_token (JWTToken): The JWT bearer token.

        Returns:
            str: The kid.
        """
        try:
            jwt_unsafe_headers: dict[str, Any] = get_unverified_header(jwt_token)
            return jwt_unsafe_headers["kid"]
        except (KeyError, InvalidTokenError) as e:
            raise InvalidJWTError("Failed to get the kid from the JWT header") from e

    @abstractmethod
    async def decode_payload(self, jwt_token: JWTToken) -> JWTBearerPayloadGeneric:
        """Decode the JWT bearer token payload.

        Args:
            jwt_token (JWTToken): The JWT bearer token.

        Returns:
            JWTBearerPayloadGeneric: The decoded JWT bearer token payload.

        Raises:
            InvalidJWTError: If the JWT bearer token is invalid.
            InvalidJWTPayploadError: If the JWT bearer token payload is invalid.
        """
        raise NotImplementedError()


GenericJWTPayload = TypeVar("GenericJWTPayload", bound=JWTPayload)


class GenericJWTBearerTokenDecoder(JWTBearerTokenDecoderAbstract[GenericJWTPayload], Generic[GenericJWTPayload]):
    """JWT bearer token classic decoder."""

    def __init__(
        self, jwt_bearer_authentication_config: JWTBearerAuthenticationConfig, jwks_store: JWKStoreAbstract
    ) -> None:
        """Initialize the JWT bearer token classic decoder.

        Args:
            jwt_bearer_authentication_config (JWTBearerAuthenticationConfig): The JWT bearer authentication
            configuration.
            jwks_store (JWKStoreAbstract): The JWKS store.
        """
        self._jwt_bearer_authentication_config: JWTBearerAuthenticationConfig = jwt_bearer_authentication_config
        self._jwks_store: JWKStoreAbstract = jwks_store
        self._payload_model: type[GenericJWTPayload] = get_args(self.__orig_bases__[0])[0]  # type: ignore[attr-defined]

    async def decode_payload(self, jwt_token: JWTToken) -> GenericJWTPayload:
        """Decode the JWT bearer token."""
        # Get the kid from the JWT header
        kid: str = self.get_kid_from_jwt_unsafe_header(jwt_token=jwt_token)
        # Get the JWK from the JWKS store
        jwk: PyJWK = await self._jwks_store.get_jwk(kid=kid)
        issuer: OAuth2Issuer = await self._jwks_store.get_issuer_by_kid(kid=kid)
        # Decode the JWT bearer token payload
        jwt_decoded: dict[str, Any] = await decode_jwt_token_payload(
            jwt_token=jwt_token,
            public_key=jwk,
            jwt_bearer_authentication_config=self._jwt_bearer_authentication_config,
            issuer=issuer,
        )
        try:
            return self._payload_model.model_validate(jwt_decoded)
        except ValidationError as e:
            raise InvalidJWTPayploadError("Failed to validate the JWT bearer token payload") from e



"""Provides security-related functions for the API."""

from .configs import (
    DependsJWTBearerAuthenticationConfig,
    JWTBearerAuthenticationConfig,
    JWTBearerAuthenticationConfigBuilder,
)
from .decoders import GenericJWTBearerTokenDecoder, JWTBearerTokenDecoderAbstract, decode_jwt_token_payload
from .exceptions import (
    ExpiredJWTError,
    InvalidJWTError,
    InvalidJWTPayploadError,
    JWTAuthenticationError,
    JWTBearerAuthenticationConfigBuilderError,
    MissingJWTCredentialsError,
    NotVerifiedJWTError,
)
from .objects import JWTPayload
from .services import JWTAuthenticationServiceAbstract
from .stores import (
    DependsHydraJWKStoreMemory,
    JWKStoreAbstract,
    JWKStoreMemory,
    configure_jwks_in_memory_store_from_hydra_introspect_services,
)
from .verifiers import GenericHydraJWTVerifier, JWTNoneVerifier, JWTVerifierAbstract

__all__: list[str] = [
    "DependsHydraJWKStoreMemory",
    "DependsJWTBearerAuthenticationConfig",
    "ExpiredJWTError",
    "GenericHydraJWTVerifier",
    "GenericJWTBearerTokenDecoder",
    "InvalidJWTError",
    "InvalidJWTPayploadError",
    "JWKStoreAbstract",
    "JWKStoreMemory",
    "JWTAuthenticationError",
    "JWTAuthenticationServiceAbstract",
    "JWTBearerAuthenticationConfig",
    "JWTBearerAuthenticationConfigBuilder",
    "JWTBearerAuthenticationConfigBuilderError",
    "JWTBearerTokenDecoderAbstract",
    "JWTNoneVerifier",
    "JWTPayload",
    "JWTVerifierAbstract",
    "MissingJWTCredentialsError",
    "NotVerifiedJWTError",
    "configure_jwks_in_memory_store_from_hydra_introspect_services",
    "decode_jwt_token_payload",
]



"""Provides the JWT bearer authentication service."""

from http import HTTPStatus
from typing import Generic, TypeVar

from fastapi import HTTPException, Request

from fastapi_factory_utilities.core.security.abstracts import AuthenticationAbstract
from fastapi_factory_utilities.core.security.types import JWTToken, OAuth2Issuer

from .configs import JWTBearerAuthenticationConfig
from .decoders import JWTBearerTokenDecoderAbstract
from .exceptions import InvalidJWTError, InvalidJWTPayploadError, MissingJWTCredentialsError, NotVerifiedJWTError
from .extraction_strategies import extract_token_from_request
from .objects import JWTPayload
from .verifiers import JWTVerifierAbstract

JWTBearerPayloadGeneric = TypeVar("JWTBearerPayloadGeneric", bound=JWTPayload)


class JWTAuthenticationServiceAbstract(AuthenticationAbstract, Generic[JWTBearerPayloadGeneric]):
    """JWT authentication service.

    This service is the orchestrator for the JWT bearer authentication.
    """

    def __init__(
        self,
        identifier: str,
        jwt_bearer_authentication_config: JWTBearerAuthenticationConfig,
        jwt_verifier: JWTVerifierAbstract[JWTBearerPayloadGeneric],
        jwt_decoder: JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric],
        raise_exception: bool = True,
    ) -> None:
        """Initialize the JWT bearer authentication service.

        Args:
            identifier (str): The identifier of the JWT bearer authentication service.
            jwt_bearer_authentication_config (JWTBearerAuthenticationConfig): The JWT bearer authentication
            configuration.
            jwt_verifier (JWTVerifierAbstract): The JWT bearer token verifier.
            jwt_decoder (JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]): The JWT bearer token decoder.
            raise_exception (bool, optional): Whether to raise an exception or return None. Defaults to True.
        """
        # Configuration and Behavior
        self._identifier: str = identifier
        self._jwt_bearer_authentication_config: JWTBearerAuthenticationConfig = jwt_bearer_authentication_config
        self._issuer: OAuth2Issuer = jwt_bearer_authentication_config.issuer
        self._jwt_verifier: JWTVerifierAbstract[JWTBearerPayloadGeneric] = jwt_verifier
        self._jwt_decoder: JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric] = jwt_decoder
        # Runtime variables
        self._jwt: JWTToken | None = None
        self._jwt_payload: JWTBearerPayloadGeneric | None = None
        super().__init__(raise_exception=raise_exception)

    @property
    def verifier(self) -> JWTVerifierAbstract[JWTBearerPayloadGeneric]:
        """Get the JWT bearer token verifier.

        Returns:
            JWTVerifierAbstract[JWTBearerPayloadGeneric]: The JWT bearer token verifier.
        """
        return self._jwt_verifier

    @property
    def decoder(self) -> JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]:
        """Get the JWT bearer token decoder.

        Returns:
            JWTBearerTokenDecoderAbstract[JWTBearerPayloadGeneric]: The JWT bearer token decoder.
        """
        return self._jwt_decoder

    @property
    def payload(self) -> JWTBearerPayloadGeneric | None:
        """Get the JWT bearer payload.

        Returns:
            JWTBearerPayloadGeneric | None: The JWT bearer payload, or None if not authenticated yet.
        """
        return self._jwt_payload

    async def authenticate(self, request: Request) -> None:
        """Authenticate the JWT bearer token.

        Args:
            request (Request): The request object.

        Returns:
            None: If the authentication is successful or not raise_exception is False.

        Raises:
            MissingJWTCredentialsError: If the authorization header is missing.
            InvalidJWTError: If the authorization header is invalid.
            InvalidJWTPayploadError: If the JWT bearer token payload is invalid.
            NotVerifiedJWTError: If the JWT bearer token is not verified.
        """
        try:
            self._jwt = extract_token_from_request(
                request=request, jwt_bearer_authentication_config=self._jwt_bearer_authentication_config
            )
        except (MissingJWTCredentialsError, InvalidJWTError) as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail=str(e)))

        try:
            self._jwt_payload = await self._jwt_decoder.decode_payload(jwt_token=self._jwt)
        except (InvalidJWTError, InvalidJWTPayploadError) as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.FORBIDDEN, detail=str(e)))

        try:
            await self._jwt_verifier.verify(jwt_token=self._jwt, jwt_payload=self._jwt_payload)
        except NotVerifiedJWTError as e:
            return self.raise_exception(HTTPException(status_code=HTTPStatus.FORBIDDEN, detail=str(e)))

        return



"""Provides the dependencies for the CSRF protect."""

from http import HTTPStatus
from typing import ClassVar

from fastapi import FastAPI, Request
from fastapi.datastructures import State
from fastapi.responses import JSONResponse
from fastapi_csrf_protect import CsrfProtect
from fastapi_csrf_protect.exceptions import CsrfProtectError
from structlog.stdlib import BoundLogger, get_logger

_logger: BoundLogger = get_logger(__package__)


class DependsCsrfProtect:
    """Dependency for the CSRF protect."""

    STATE_KEY: ClassVar[str] = "csrf_protect"

    @classmethod
    def import_to_state(cls, state: State, csrf_protect: CsrfProtect) -> None:
        """Import the CSRF protect to the state."""
        state[cls.STATE_KEY] = csrf_protect

    @classmethod
    def export_from_state(cls, state: State) -> CsrfProtect:
        """Export the CSRF protect from the state."""
        return state[cls.STATE_KEY]

    def __call__(self, request: Request) -> CsrfProtect:
        """Call the dependency."""
        return self.export_from_state(state=request.app.state)


depends_csrf_protect: DependsCsrfProtect = DependsCsrfProtect()


async def csrf_protect_exception_handler(request: Request, exc: Exception) -> JSONResponse:
    """CSRF exception handler."""
    _logger.warning("CSRF error", request=request, exc=exc)
    return JSONResponse(status_code=HTTPStatus.FORBIDDEN, content={"detail": "CSRF token is invalid"})


def register_csrf_protect_exception_handler(app: FastAPI) -> None:
    """Register the CSRF exception handler."""
    app.add_exception_handler(CsrfProtectError, csrf_protect_exception_handler)



"""Mocker for ODM plugin.

Objectives:
- Provide an implementation in memory for the Repository Class.
- This RepositoryInMemory must be a subclass of the AbstractRepository class.
- This RepositoryInMemory must be a singleton with all the data in memory.
- This RepositoryInMemory must accept generic like the real one.
"""

import datetime
import re
from abc import ABC
from collections.abc import AsyncGenerator, Callable, Mapping
from contextlib import asynccontextmanager
from copy import deepcopy
from functools import wraps
from typing import Any, Generic, TypeVar, get_args
from uuid import UUID

from beanie import SortDirection

from fastapi_factory_utilities.core.plugins.odm_plugin.documents import BaseDocument
from fastapi_factory_utilities.core.plugins.odm_plugin.exceptions import (
    OperationError,
    UnableToCreateEntityDueToDuplicateKeyError,
)
from fastapi_factory_utilities.core.plugins.odm_plugin.helpers import PersistedEntity

DocumentGenericType = TypeVar("DocumentGenericType", bound=BaseDocument)  # pylint: disable=invalid-name
EntityGenericType = TypeVar("EntityGenericType", bound=PersistedEntity[UUID])  # pylint: disable=invalid-name


class MockQueryField(str):
    """Mock field for query expressions without Beanie initialization.

    This allows Document fields to be accessed for query expressions
    (e.g., Document.field == value) without requiring full Beanie initialization.

    Inherits from str to be hashable and compatible with Beanie operators.
    """

    def __new__(cls, field_name: str) -> "MockQueryField":
        """Create a new MockQueryField instance.

        Args:
            field_name: The name of the field.

        Returns:
            A new MockQueryField instance.
        """
        instance = str.__new__(cls, field_name)
        return instance

    @property
    def field_name(self) -> str:
        """Get the field name.

        Returns:
            The field name as a string.
        """
        return str(self)

    def __hash__(self) -> int:
        """Return hash of the field name.

        Returns:
            Hash of the field name string.
        """
        return hash(str(self))

    def __eq__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support equality comparison (field == value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter that can be used by the in-memory repository.
        """
        return {self.field_name: other}

    def __ne__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support inequality comparison (field != value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter representation.
        """
        return {self.field_name: {"$ne": other}}

    def __lt__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support less than comparison (field < value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter representation.
        """
        return {self.field_name: {"$lt": other}}

    def __le__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support less than or equal comparison (field <= value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter representation.
        """
        return {self.field_name: {"$lte": other}}

    def __gt__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support greater than comparison (field > value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter representation.
        """
        return {self.field_name: {"$gt": other}}

    def __ge__(self, other: Any) -> dict[str, Any]:  # type: ignore[override]
        """Support greater than or equal comparison (field >= value).

        Args:
            other: The value to compare against.

        Returns:
            A dictionary filter representation.
        """
        return {self.field_name: {"$gte": other}}


def _setup_document_fields_for_mock(document_type: type[BaseDocument]) -> None:
    """Set up document fields to support query expressions without Beanie initialization.

    This function adds MockQueryField descriptors to the document class for each field,
    allowing query expressions like `Document.field == value` to work without requiring
    full Beanie initialization.

    Args:
        document_type: The document class to set up.
    """
    # Get all fields from the Pydantic model
    for field_name in document_type.model_fields:
        # Only set up fields that don't already exist as class attributes
        # (to avoid overriding Beanie's setup if it's already done)
        if not hasattr(document_type, field_name) or isinstance(getattr(document_type, field_name), property):
            setattr(document_type, field_name, MockQueryField(field_name))


def managed_session() -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to manage the session.

    It will introspect the function arguments and check if the session is passed as a keyword argument.
    If it is not, it will create a new session and pass it to the function.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            if "session" in kwargs:
                return await func(*args, **kwargs)

            async with args[0].get_session() as session:
                return await func(*args, **kwargs, session=session)

        return wrapper

    return decorator


class AbstractRepositoryInMemory(ABC, Generic[DocumentGenericType, EntityGenericType]):
    """Abstract repository in memory for testing purposes.

    This class provides an in-memory implementation of the repository pattern,
    allowing tests to run without requiring a real database connection.
    """

    def __init__(self, entities: list[EntityGenericType] | None = None) -> None:
        """Initialize the repository.

        Args:
            entities: Optional list of entities to pre-populate the repository with.
        """
        self._entities: dict[UUID, EntityGenericType] = {}
        if entities is not None:
            for entity in entities:
                self._entities[entity.id] = entity
        # Retrieve the generic concrete types
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._document_type: type[DocumentGenericType] = generic_args[0]
        self._entity_type: type[EntityGenericType] = generic_args[1]

        # Set up document fields to support query expressions without Beanie initialization
        _setup_document_fields_for_mock(self._document_type)

    def _matches_filter(  # noqa: PLR0911, PLR0912
        self, entity: EntityGenericType, filter_dict: Mapping[str, Any]
    ) -> bool:
        """Check if an entity matches a filter dictionary.

        Supports both simple equality filters and MongoDB-style operator filters.

        Args:
            entity: The entity to check.
            filter_dict: The filter dictionary (e.g., {"field": value} or {"field": {"$lt": 5}}).

        Returns:
            True if the entity matches all filter conditions, False otherwise.
        """
        for key, value in filter_dict.items():
            entity_value: Any = getattr(entity, key)

            # Check if value is a dict with MongoDB operators
            if isinstance(value, dict):
                for op_key, op_value in value.items():
                    operator: str = str(op_key)
                    if operator == "$ne":
                        if entity_value == op_value:
                            return False
                    elif operator == "$lt":
                        if entity_value >= op_value:
                            return False
                    elif operator == "$lte":
                        if entity_value > op_value:
                            return False
                    elif operator == "$gt":
                        if entity_value <= op_value:
                            return False
                    elif operator == "$gte":
                        if entity_value < op_value:
                            return False
                    elif operator == "$in":
                        if entity_value not in op_value:
                            return False
                    elif operator == "$nin":
                        if entity_value in op_value:
                            return False
                    elif operator == "$regex":
                        # Get options from the filter dict (might be in same dict)
                        options_str: str = (
                            filter_dict.get(key, {}).get("$options", "")
                            if isinstance(filter_dict.get(key), dict)
                            else ""
                        )
                        flags: int = 0
                        if "i" in options_str:
                            flags |= re.IGNORECASE
                        if "m" in options_str:
                            flags |= re.MULTILINE
                        if "s" in options_str:
                            flags |= re.DOTALL
                        pattern = re.compile(str(op_value), flags)
                        if not pattern.search(str(entity_value)):
                            return False
                    elif operator == "$exists":
                        # Check if field exists (not None) or doesn't exist (is None)
                        field_exists = entity_value is not None
                        if op_value != field_exists:
                            return False
                    elif operator == "$all":
                        # All values must be in the entity value (which should be a list/sequence)
                        if not isinstance(entity_value, (list, tuple, set)):
                            return False
                        entity_set: set[Any] = set(entity_value) if not isinstance(entity_value, set) else entity_value
                        if not all(item in entity_set for item in op_value):
                            return False
                    elif operator == "$size":
                        # Check the size of the entity value (which should be a list/sequence)
                        if not hasattr(entity_value, "__len__"):
                            return False
                        if len(entity_value) != op_value:
                            return False
                    elif operator == "$options":
                        # Options is handled together with $regex
                        continue
                    else:
                        # Unknown operator, treat as inequality
                        return False
            # Simple equality check
            elif entity_value != value:
                return False

        return True

    @asynccontextmanager
    async def get_session(self) -> AsyncGenerator[None, None]:
        """Get a session context manager.

        Yields:
            None: No actual session is needed for in-memory operations.
        """
        yield

    @managed_session()
    async def insert(self, entity: EntityGenericType, session: None = None) -> EntityGenericType:  # pylint: disable=unused-argument
        """Insert an entity into the repository.

        Args:
            entity: The entity to insert.
            session: The session to use (unused in memory implementation).

        Returns:
            The created entity with timestamps set.

        Raises:
            UnableToCreateEntityDueToDuplicateKeyError: If an entity with the same ID already exists.
        """
        insert_time: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        entity_dump: dict[str, Any] = entity.model_dump()
        entity_dump["created_at"] = insert_time
        entity_dump["updated_at"] = insert_time
        entity_created: EntityGenericType = self._entity_type(**entity_dump)

        if entity_created.id in self._entities:
            raise UnableToCreateEntityDueToDuplicateKeyError(f"Entity with ID {entity_created.id} already exists")

        self._entities[entity_created.id] = entity_created
        return entity_created

    @managed_session()
    async def update(self, entity: EntityGenericType, session: None = None) -> EntityGenericType:  # pylint: disable=unused-argument
        """Update an entity in the repository.

        Args:
            entity: The entity to update.
            session: The session to use (unused in memory implementation).

        Returns:
            The updated entity with updated_at timestamp refreshed.

        Raises:
            OperationError: If the entity does not exist in the repository.
        """
        update_time: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        entity_dump: dict[str, Any] = entity.model_dump()
        entity_dump["updated_at"] = update_time
        entity_updated: EntityGenericType = self._entity_type(**entity_dump)

        if entity_updated.id not in self._entities:
            raise OperationError(f"Entity with ID {entity_updated.id} not found")

        self._entities[entity_updated.id] = entity_updated
        return entity_updated

    @managed_session()
    async def get_one_by_id(self, entity_id: UUID, session: None = None) -> EntityGenericType | None:  # pylint: disable=unused-argument
        """Get an entity by its ID.

        Args:
            entity_id: The ID of the entity to retrieve.
            session: The session to use (unused in memory implementation).

        Returns:
            The entity if found, None otherwise.
        """
        return self._entities.get(entity_id, None)

    @managed_session()
    async def delete_one_by_id(self, entity_id: UUID, raise_if_not_found: bool = False, session: None = None) -> None:  # pylint: disable=unused-argument
        """Delete an entity by its ID.

        Args:
            entity_id: The ID of the entity to delete.
            raise_if_not_found: If True, raises OperationError when entity is not found.
            session: The session to use (unused in memory implementation).

        Raises:
            OperationError: If the entity is not found and raise_if_not_found is True.
        """
        if entity_id not in self._entities:
            if raise_if_not_found:
                raise OperationError(f"Entity with ID {entity_id} not found")
            return
        self._entities.pop(entity_id)

    @managed_session()
    async def find(  # noqa: PLR0913  # pylint: disable=unused-argument
        self,
        *args: Any,
        projection_model: None = None,
        skip: int | None = None,
        limit: int | None = None,
        sort: None | str | list[tuple[str, SortDirection]] = None,
        session: None = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        lazy_parse: bool = False,
        nesting_depth: int | None = None,
        nesting_depths_per_field: dict[str, int] | None = None,
        **pymongo_kwargs: Any,
    ) -> list[EntityGenericType]:
        """Find entities in the repository.

        Args:
            *args: Filter arguments. Can be Mapping for field filters, bool for boolean filters,
                   or any other type (treated as boolean).
            projection_model: Unused in memory implementation.
            skip: Number of entities to skip.
            limit: Maximum number of entities to return.
            sort: Sort order as a list of tuples (field_name, SortDirection).
            session: The session to use (unused in memory implementation).
            ignore_cache: Unused in memory implementation.
            fetch_links: Unused in memory implementation.
            lazy_parse: Unused in memory implementation.
            nesting_depth: Unused in memory implementation.
            nesting_depths_per_field: Unused in memory implementation.
            **pymongo_kwargs: Unused in memory implementation.

        Returns:
            A list of entities matching the filters, sorted, skipped, and limited as specified.
            Returns deep copies to prevent accidental modification of stored entities.
        """
        initial_list: list[EntityGenericType] = deepcopy(list(self._entities.values()))

        # Apply the filters
        if args:
            for arg in args:
                if isinstance(arg, Mapping):
                    # Dictionary filter or Beanie operator (both are Mappings)
                    # Supports: {"field": value}, {"field": {"$op": value}},
                    # or Beanie operators like In(), NotIn(), etc.
                    filter_arg: Mapping[str, Any] = arg
                    initial_list = [entity for entity in initial_list if self._matches_filter(entity, filter_arg)]
                elif isinstance(arg, bool):
                    # Boolean filter: if False, filter out all entities
                    initial_list = [entity for entity in initial_list if arg]
                else:
                    # Treat any other type (including Beanie query expressions) as boolean
                    # This handles cases where query expressions evaluate to bool
                    initial_list = [entity for entity in initial_list if arg]

        # Apply the sorting
        if sort:
            initial_list = sorted(
                initial_list, key=lambda x: x.model_dump()[sort[0][0]], reverse=sort[0][1] == SortDirection.DESCENDING
            )

        # Apply the skip
        if skip:
            initial_list = initial_list[skip:]

        # Apply the limit
        if limit:
            initial_list = initial_list[:limit]

        return initial_list



"""Builder for the Aiohttp client."""

from typing import Self

from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol

from .configs import HttpServiceDependencyConfig
from .factories import build_http_dependency_config
from .resources import AioHttpClientResource


class AioHttpClientBuilder:
    """Builder for the Aiohttp client."""

    def __init__(
        self,
        keys: list[str],
        application: ApplicationAbstractProtocol,
    ) -> None:
        """Initialize the Aiohttp client builder."""
        self._application: ApplicationAbstractProtocol = application
        self._keys: list[str] = keys
        self._configs: dict[str, HttpServiceDependencyConfig] = {}
        self._resources: dict[str, AioHttpClientResource] = {}

    def build_configs(self) -> Self:
        """Build the HTTP dependency configs."""
        if self._application.PACKAGE_NAME == "":
            raise ValueError("The application package name is not set")

        for key in self._keys:
            self._configs[key] = build_http_dependency_config(
                key=key, application_package=self._application.PACKAGE_NAME
            )
        return self

    def build_resources(self) -> Self:
        """Build the Aiohttp client."""
        for key, config in self._configs.items():
            self._resources[key] = AioHttpClientResource(dependency_config=config)
        return self

    @property
    def resources(self) -> dict[str, AioHttpClientResource]:
        """Get the Aiohttp client resources."""
        return self._resources



"""Http service dependency config."""

from pydantic import BaseModel, Field, HttpUrl


class HttpServiceDependencyConfig(BaseModel):
    """Http service dependency config."""

    url: HttpUrl | None = Field(default=None, description="Base Url used to build the client session")
    # Pool configuration
    limit: int = Field(default=10, description="Limit of connections")
    limit_per_host: int = Field(default=10, description="Limit of connections per host")
    # DNS cache configuration
    use_dns_cache: bool = Field(default=True, description="Use DNS cache")
    ttl_dns_cache: int = Field(default=300, description="TTL DNS cache in seconds")
    verify_ssl: bool = Field(default=True, description="Verify SSL")
    # SSL configuration to verify the server certificate
    ssl_ca_path: str | None = Field(default=None, description="SSL CA path")
    # SSL configuration for the client use certificate authentication
    ssl_certfile: str | None = Field(default=None, description="SSL certificate file")
    ssl_keyfile: str | None = Field(default=None, description="SSL key file")
    ssl_keyfile_password: str | None = Field(default=None, description="SSL key file password")
    # Application Graceful shutdown configuration
    graceful_shutdown_timeout: int = Field(default=10, description="Graceful shutdown timeout in seconds")



"""Provides services for books."""

from uuid import UUID

from fastapi import Request
from opentelemetry import metrics

from fastapi_factory_utilities.core.plugins.odm_plugin.depends import (
    depends_odm_database,
)
from fastapi_factory_utilities.core.plugins.opentelemetry_plugin.helpers import (
    trace_span,
)
from fastapi_factory_utilities.example.entities.books import BookEntity
from fastapi_factory_utilities.example.models.books.repository import BookRepository


class BookService:
    """Provides services for books."""

    # Metrics Definitions
    METER_COUNTER_BOOK_GET_NAME: str = "book_get"
    METER_COUNTER_BOOK_ADD_NAME: str = "book_add"
    METER_COUNTER_BOOK_REMOVE_NAME: str = "book_remove"
    METER_COUNTER_BOOK_UPDATE_NAME: str = "book_update"
    # ====================

    meter: metrics.Meter = metrics.get_meter(__name__)

    METER_COUNTER_BOOK_GET: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_GET_NAME, description="The number of books retrieved."
    )
    METER_COUNTER_BOOK_ADD: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_ADD_NAME,
        description="The number of books added.",
    )
    METER_COUNTER_BOOK_REMOVE: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_REMOVE_NAME,
        description="The number of books removed.",
    )
    METER_COUNTER_BOOK_UPDATE: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_UPDATE_NAME,
        description="The number of books updated.",
    )

    def __init__(
        self,
        book_repository: BookRepository,
    ) -> None:
        """Initialize the service.

        Args:
            book_repository: The book repository.
        """
        self.book_repository: BookRepository = book_repository

    @trace_span(name="Add Book")
    async def add_book(self, book: BookEntity) -> BookEntity:
        """Add a book.

        Args:
            book: The book to add.

        Returns:
            The created book entity.

        Raises:
            UnableToCreateEntityDueToDuplicateKeyError: If a book with the same title already exists.
            OperationError: If the operation fails.
        """
        created_book: BookEntity = await self.book_repository.insert(entity=book)
        self.METER_COUNTER_BOOK_ADD.add(amount=1)
        return created_book

    async def get_book(self, book_id: UUID) -> BookEntity:
        """Get a book.

        Args:
            book_id: The book id.

        Returns:
            The book entity.

        Raises:
            ValueError: If the book does not exist.
            OperationError: If the operation fails.
        """
        book: BookEntity | None = await self.book_repository.get_one_by_id(entity_id=book_id)

        if book is None:
            raise ValueError(f"Book with id {book_id} does not exist.")

        self.METER_COUNTER_BOOK_GET.add(amount=1, attributes={"book_count": 1})
        return book

    async def get_all_books(self) -> list[BookEntity]:
        """Get all books.

        Returns:
            All books.

        Raises:
            OperationError: If the operation fails.
        """
        books: list[BookEntity] = await self.book_repository.find()
        self.METER_COUNTER_BOOK_GET.add(amount=1, attributes={"book_count": len(books)})
        return books

    @trace_span(name="Remove Book")
    async def remove_book(self, book_id: UUID) -> None:
        """Remove a book.

        Args:
            book_id: The book id.

        Raises:
            ValueError: If the book does not exist.
            OperationError: If the operation fails.
        """
        await self.book_repository.delete_one_by_id(entity_id=book_id, raise_if_not_found=True)

        self.METER_COUNTER_BOOK_REMOVE.add(amount=1)

    @trace_span(name="Update Book")
    async def update_book(self, book: BookEntity) -> BookEntity:
        """Update a book.

        Args:
            book: The book to update.

        Returns:
            The updated book entity.

        Raises:
            ValueError: If the book does not exist.
            OperationError: If the operation fails.
        """
        # Check if book exists
        existing_book: BookEntity | None = await self.book_repository.get_one_by_id(entity_id=book.id)

        if existing_book is None:
            raise ValueError(f"Book with id {book.id} does not exist.")

        updated_book: BookEntity = await self.book_repository.update(entity=book)
        self.METER_COUNTER_BOOK_UPDATE.add(amount=1)
        return updated_book


def depends_book_service(request: Request) -> BookService:
    """Provide Book Service.

    Args:
        request: FastAPI request object.

    Returns:
        BookService instance.
    """
    return BookService(book_repository=BookRepository(database=depends_odm_database(request=request)))



"""Provides utilities for pagination."""

from fastapi import Request

from .types import PaginationPageOffset, PaginationSize


def resolve_offset(page_offset: PaginationPageOffset, page_size: PaginationSize) -> int:
    """Resolve the offset from the page offset and page size."""
    return page_offset * page_size


def depends_pagination_page_offset(request: Request) -> PaginationPageOffset:
    """Get the pagination page offset from the request.

    Args:
        request (Request): The request.

    Returns:
        PaginationPageOffset: The pagination page offset.

    Raises:
        ValidationError: If the pagination page offset is not a valid integer.
    """
    raw_page_offset: str | None = request.query_params.get("page_offset")
    if raw_page_offset is None:
        return PaginationPageOffset.default()

    return PaginationPageOffset(int(raw_page_offset))


def depends_pagination_page_size(request: Request) -> PaginationSize:
    """Get the pagination page size from the request.

    Args:
        request (Request): The request.

    Returns:
        PaginationSize: The pagination page size.

    Raises:
        ValidationError: If the pagination page size is not a valid integer.
    """
    raw_page_size: str | None = request.query_params.get("page_size")
    if raw_page_size is None:
        return PaginationSize.default()
    return PaginationSize(int(raw_page_size))


__all__: list[str] = [
    "PaginationPageOffset",
    "PaginationSize",
    "depends_pagination_page_offset",
    "depends_pagination_page_size",
    "resolve_offset",
]



"""Helper functions for ODM plugins."""

import datetime
import uuid
from collections.abc import Callable
from typing import Generic, TypeVar, cast

from pydantic import BaseModel, Field

GenericPersistedEntityId = TypeVar("GenericPersistedEntityId", bound=uuid.UUID)


class PersistedEntity(BaseModel, Generic[GenericPersistedEntityId]):
    """Base class for persisted entities.

    Attributes:
        id: The ID of the persisted entity.
        revision_id: The revision ID of the persisted entity.
        created_at: The creation date of the persisted entity.
        updated_at: The last update date of the persisted entity.

    ```python
    from fastapi_factory_utilities.core.plugins.odm_plugin.helpers import PersistedEntity

    BookEntityId = NewType("BookEntityId", uuid.UUID)


    class BookEntity(PersistedEntity[BookEntityId]):
        title: str = Field(title="Title of the book")
        author: str = Field(title="Author name")
        published_year: int = Field(title="Year published")
    ```
    """

    id: GenericPersistedEntityId = Field(default_factory=cast(Callable[[], GenericPersistedEntityId], uuid.uuid4))

    revision_id: uuid.UUID | None = Field(default=None)
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    updated_at: datetime.datetime = Field(default_factory=datetime.datetime.now)



"""Kratos service module."""

from .enums import (
    AuthenticationMethodEnum,
    AuthenticatorAssuranceLevelEnum,
    KratosFlowTypeEnum,
    KratosIdentityPatchOpEnum,
    KratosIdentityStateEnum,
)
from .exceptions import KratosIdentityNotFoundError, KratosOperationError, KratosSessionInvalidError
from .objects import (
    KratosIdentityObject,
    KratosRecoveryAddressObject,
    KratosSessionObject,
    KratosTraitsObject,
    MetadataObject,
)
from .services import KratosGenericWhoamiService, KratosIdentityGenericService, KratosIdentityPatchObject
from .types import (
    KratosExternalId,
    KratosIdentityId,
    KratosProvider,
    KratosRecoveryCode,
    KratosRecoveryLink,
    KratosSchemaId,
)

__all__: list[str] = [
    "AuthenticationMethodEnum",
    "AuthenticatorAssuranceLevelEnum",
    "KratosExternalId",
    "KratosFlowTypeEnum",
    "KratosGenericWhoamiService",
    "KratosIdentityGenericService",
    "KratosIdentityId",
    "KratosIdentityNotFoundError",
    "KratosIdentityObject",
    "KratosIdentityPatchObject",
    "KratosIdentityPatchOpEnum",
    "KratosIdentityStateEnum",
    "KratosOperationError",
    "KratosProvider",
    "KratosRecoveryAddressObject",
    "KratosRecoveryCode",
    "KratosRecoveryLink",
    "KratosSchemaId",
    "KratosSessionInvalidError",
    "KratosSessionObject",
    "KratosTraitsObject",
    "MetadataObject",
]



"""Provides utility functions for Ory."""

import re
from urllib.parse import parse_qs, urlparse


def get_next_page_token_from_link_header(link_header: str | None) -> str | None:
    """Parse the Link header and extract the next page token.

    Args:
        link_header (str | None): The Link header value from the HTTP response.

    Returns:
        str | None: The next page token if found, None otherwise.

    Example:
        >>> link_header = (
        ...     '</admin/clients?page_size=5&page_token=euKoY1BqY3J8GVax>; rel="first",'
        ...     '</admin/clients?page_size=5&page_token=QLux4Tu5gb8JfW70>; rel="next"'
        ... )
        >>> token = KratosIdentityGenericService._parse_link_header(link_header)
        >>> print(token)
        'QLux4Tu5gb8JfW70'
    """
    if link_header is None:
        return None

    # Parse Link header: <url>; rel="type", <url>; rel="type"
    # Find all links with rel="next"
    pattern: re.Pattern[str] = re.compile(r'<([^>]+)>;\s*rel="next"')
    matches: list[str] = pattern.findall(link_header)

    if not matches:
        return None

    # Get the first match (should only be one "next" link)
    next_url: str = matches[0]

    # Parse the URL to extract query parameters
    parsed_url = urlparse(next_url)
    query_params: dict[str, list[str]] = parse_qs(parsed_url.query)

    # Extract page_token from query parameters
    page_tokens: list[str] | None = query_params.get("page_token")
    if page_tokens and len(page_tokens) > 0:
        return page_tokens[0]

    return None



"""Provides the security types."""

from typing import NewType

JWTToken = NewType("JWTToken", str)
OAuth2Scope = NewType("OAuth2Scope", str)
OAuth2Audience = NewType("OAuth2Audience", str)
OAuth2Issuer = NewType("OAuth2Issuer", str)
OAuth2Subject = NewType("OAuth2Subject", str)

__all__: list[str] = [
    "JWTToken",
    "OAuth2Audience",
    "OAuth2Issuer",
    "OAuth2Scope",
    "OAuth2Subject",
]



"""Provides the Book entity."""

from pydantic import Field

from fastapi_factory_utilities.core.plugins.odm_plugin.helpers import PersistedEntity

from .enums import BookType
from .types import BookEntityId, BookName


class BookEntity(PersistedEntity[BookEntityId]):
    """Book entity."""

    title: BookName = Field(title="Title of the book")
    book_type: BookType = Field(title="Type of book")



"""Provides the security authentication abstract classes."""

from abc import ABC, abstractmethod

from fastapi import Request


class AuthenticationAbstract(ABC):
    """Authentication abstract class."""

    def __init__(self, raise_exception: bool = True) -> None:
        """Initialize the authentication abstract class.

        Args:
            raise_exception (bool): Whether to raise an exception or return None.
        """
        self._raise_exception: bool = raise_exception
        self._errors: list[Exception] = []

    def has_errors(self) -> bool:
        """Check if the authentication has errors.

        Returns:
            bool: True if the authentication has errors, False otherwise.
        """
        return len(self._errors) > 0

    def raise_exception(self, exception: Exception) -> None:
        """Raise the exception if the authentication has errors.

        Args:
            exception (Exception): The exception to raise.
        """
        if self._raise_exception:
            raise exception
        else:
            self._errors.append(exception)

    @abstractmethod
    async def authenticate(self, request: Request) -> None:
        """Authenticate the request."""
        raise NotImplementedError()



"""Provides the objects for the audit service."""

import datetime
import uuid
from typing import Any, Generic, NewType, TypeVar, cast

from pydantic import BaseModel, PrivateAttr, field_validator

EntityName = NewType("EntityName", str)
EntityFunctionalEventName = NewType("EntityFunctionalEventName", str)
ServiceName = NewType("ServiceName", str)


class AuditableEntity(BaseModel):
    """Auditable entity.

    Attributes:
        _audit_name: The name of the audit. This is used to identify the entity in the audit trail.
            It's must be unique on the whole ecosystem.
        id: The ID of the entity.
        created_at: The creation date of the entity.
        updated_at: The last update date of the entity.
        deleted_at: The deletion date of the entity.
    """

    _audit_name: EntityName = PrivateAttr(default=EntityName(""))

    id: uuid.UUID
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deleted_at: datetime.datetime | None = None

    def get_audit_name(self) -> str:
        """Get the audit name."""
        if self._audit_name == EntityName(""):
            raise ValueError("Audit name is not set.")
        return self._audit_name


AuditEventActorGeneric = TypeVar("AuditEventActorGeneric", bound=AuditableEntity)


class AuditEventObject(BaseModel, Generic[AuditEventActorGeneric]):
    """Audit event object.

    Attributes:
        what: The name of the entity.
        why: The name of the functional event.
        where: The name of the service.
        when: The date and time of the event.
        who: The dictionary of ids of the actors involved in the event.
        (always at least contains the id of the actor who performed the event but can contains more ids id needed like
        segmentation ids (realms, groups, etc.))
    """

    what: EntityName
    why: EntityFunctionalEventName
    where: ServiceName
    when: datetime.datetime
    who: dict[str, Any]

    @field_validator("who")
    @classmethod
    def who_validator(cls, value: Any) -> dict[str, Any]:
        """Validate the who."""
        # Check if the value is a dictionary
        if not isinstance(value, dict):
            raise ValueError("Who must be a dictionary.")
        value_dict: dict[str, Any] = cast(dict[str, Any], value)
        # Check if the dictionary is empty
        if len(value_dict) == 0:
            raise ValueError("Who must not be empty.")
        # Check if the dictionary contains id key
        if "id" not in value_dict:
            raise ValueError("Who must contain id key.")
        return value_dict



"""Provides the dependencies for the application."""

from typing import Generic, TypeVar, cast

from fastapi import Depends, Request
from fastapi.datastructures import State

from .config import AppCsrfConfig, BaseApplicationConfig, RootConfig

GenericRootConfig = TypeVar("GenericRootConfig", bound=RootConfig)
GenericApplicationConfig = TypeVar("GenericApplicationConfig", bound=BaseApplicationConfig)


class DependsRootConfig(Generic[GenericRootConfig]):
    """Dependency for the root config."""

    @classmethod
    def export_from_state(cls, state: State) -> GenericRootConfig:
        """Export the root config from the state."""
        return getattr(state, "config")

    @classmethod
    def import_to_state(cls, state: State, config: GenericRootConfig) -> None:
        """Import the root config to the state."""
        setattr(state, "config", config)

    def __call__(self, request: Request) -> GenericRootConfig:
        """Dependency for the root config."""
        if not hasattr(request.app.state, "config"):
            raise ValueError("Root config not found in the state.")
        return self.export_from_state(state=request.app.state)


depends_root_config = DependsRootConfig[RootConfig]()


class DependsApplicationConfig(Generic[GenericApplicationConfig]):
    """Dependency for the application config."""

    def __call__(self, root_config: RootConfig = Depends(depends_root_config)) -> GenericApplicationConfig:
        """Dependency for the application config."""
        return cast(GenericApplicationConfig, root_config.application)


depends_application_config = DependsApplicationConfig[BaseApplicationConfig]()


def depends_csrf_config(root_config: RootConfig = Depends(depends_root_config)) -> AppCsrfConfig:
    """Dependency for the CSRF config."""
    if root_config.csrf is None:
        raise ValueError("CSRF config not found in the root config.")
    return root_config.csrf



"""Provides the configurations for the JWT bearer token."""

from enum import StrEnum
from typing import ClassVar, Self

from fastapi import Request
from fastapi.datastructures import State
from jwt.algorithms import get_default_algorithms, requires_cryptography
from pydantic import BaseModel, ConfigDict, Field, field_validator

from fastapi_factory_utilities.core.security.types import OAuth2Issuer
from fastapi_factory_utilities.core.utils.configs import (
    UnableToReadConfigFileError,
    ValueErrorConfigError,
    build_config_from_file_in_package,
)

from .exceptions import JWTBearerAuthenticationConfigBuilderError


class JWTLocation(StrEnum):
    """JWT location."""

    HEADER = "header"
    AUTHORIZATION_BEARER = "authorization_bearer"
    COOKIE = "cookie"


class JWTBearerAuthenticationConfig(BaseModel):
    """JWT bearer token authentication configuration."""

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    authorized_algorithms: list[str] = Field(
        default_factory=lambda: list(get_default_algorithms().keys()), description="The authorized algorithms."
    )

    authorized_audiences: list[str] | None = Field(default=None, description="The authorized audiences.")
    issuer: OAuth2Issuer = Field(description="The authorized issuers.")
    audience: str | None = Field(default=None, description="The audience.")

    # JWT location
    authorized_locations: list[JWTLocation] = Field(
        default_factory=lambda: [JWTLocation.AUTHORIZATION_BEARER], description="The authorized locations."
    )
    header_name: str | None = Field(default=None, description="The header name.")
    cookie_name: str | None = Field(default=None, description="The cookie name.")

    @field_validator("authorized_audiences", mode="before")
    @classmethod
    def validate_authorized_audiences(cls, v: str | list[str]) -> list[str]:
        """Validate the authorized audiences.

        Example:
            "aud1,aud2,aud3" -> ["aud1", "aud2", "aud3"]
            ["aud1", "aud2", "aud3"] -> ["aud1", "aud2", "aud3"]
        """
        if isinstance(v, str):
            v = v.split(sep=",")
        v = [item.strip() for item in v if item.strip()]
        if len(v) == 0:
            raise ValueError("Invalid value: empty list after processing")
        return list(set(v))

    @field_validator("authorized_algorithms")
    @classmethod
    def validate_authorized_algorithms(cls, v: list[str]) -> list[str]:
        """Validate the authorized algorithms."""
        invalid_algorithms: list[str] = []
        for algorithm in v:
            if algorithm not in requires_cryptography:
                invalid_algorithms.append(algorithm)
        if invalid_algorithms:
            raise ValueError(f"Invalid algorithms: {invalid_algorithms}")
        return v


class JWTBearerAuthenticationConfigBuilder:
    """Builder for the JWT bearer authentication configuration.

    The builder will build the JWT bearer authentication configuration from the application YAML file.
    It must be used in the configuration hook of the application.
    """

    APPLICATION_YAML_BASE_JWT_CONFIG_KEY: str = "jwt_configs"

    def __init__(self, key: str) -> None:
        """Initialize the JWT bearer authentication configuration builder.

        Args:
            key: The key of the JWT bearer authentication configuration.
        """
        if not key or len(key) == 0:
            raise JWTBearerAuthenticationConfigBuilderError("Key cannot be empty")

        self._key: str = key
        self._config_key: str = f"{self.APPLICATION_YAML_BASE_JWT_CONFIG_KEY}.{self._key}"
        self._config: JWTBearerAuthenticationConfig | None = None
        self._package_name: str | None = None
        self._filename: str | None = None

    def add_application_yaml_path(self, package_name: str, filename: str) -> Self:
        """Add the application YAML path to the builder."""
        self._package_name = package_name
        self._filename = filename
        return self

    def _build_config_from_application_yaml(self) -> JWTBearerAuthenticationConfig:
        """Build the JWT bearer authentication configuration from the application YAML."""
        if self._package_name is None or self._filename is None:
            raise JWTBearerAuthenticationConfigBuilderError("Package name and filename must be provided")

        try:
            return build_config_from_file_in_package(
                package_name=self._package_name,
                filename=self._filename,
                config_class=JWTBearerAuthenticationConfig,
                yaml_base_key=self._config_key,
            )
        except (UnableToReadConfigFileError, ValueErrorConfigError) as error:
            raise JWTBearerAuthenticationConfigBuilderError("Failed to read the application YAML file") from error

    def build(self) -> JWTBearerAuthenticationConfig:
        """Build the JWT bearer authentication configuration.

        Returns:
            JWTBearerAuthenticationConfig: The built JWT bearer authentication configuration.

        Raises:
            JWTBearerAuthenticationConfigBuilderError: When no configuration is provided
            and the application YAML path is not provided.
        """
        if self._config is None and self._package_name is not None and self._filename is not None:
            self._config = self._build_config_from_application_yaml()

        if self._config is None:
            raise JWTBearerAuthenticationConfigBuilderError(
                "Neither a JWT bearer authentication configuration nor an application YAML path was provided"
            )

        return self._config

    @property
    def config(self) -> JWTBearerAuthenticationConfig:
        """Get the JWT bearer authentication configuration."""
        if self._config is None:
            raise JWTBearerAuthenticationConfigBuilderError("No configuration found")
        return self._config


class DependsJWTBearerAuthenticationConfig:
    """Dependency for the JWT bearer authentication configuration."""

    STATE_PREFIX_KEY: str = "jwt_configs"

    def __init__(self, key: str) -> None:
        """Initialize the dependency for the JWT bearer authentication configuration."""
        self._key: str = key

    @classmethod
    def export_from_state(cls, state: State, key: str) -> JWTBearerAuthenticationConfig:
        """Export the JWT bearer authentication configuration from the state."""
        config: JWTBearerAuthenticationConfig | None = getattr(state, f"{cls.STATE_PREFIX_KEY}.{key}", None)
        if config is None:
            raise JWTBearerAuthenticationConfigBuilderError(
                "JWT bearer authentication configuration not found in the state"
            )
        return config

    @classmethod
    def import_to_state(cls, state: State, config: JWTBearerAuthenticationConfig, key: str) -> None:
        """Import the JWT bearer authentication configuration to the state."""
        setattr(state, f"{cls.STATE_PREFIX_KEY}.{key}", config)

    def __call__(self, request: Request) -> JWTBearerAuthenticationConfig:
        """Dependency for the JWT bearer authentication configuration."""
        return self.export_from_state(state=request.app.state, key=self._key)



"""Provides the services for the audit service."""

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from fastapi_factory_utilities.core.plugins.aiopika import (
    AbstractPublisher,
    AiopikaPluginBaseError,
    GenericMessage,
    RoutingKey,
)

from .exceptions import AuditServiceError
from .objects import AuditEventObject, ServiceName

AuditEventGeneric = TypeVar("AuditEventGeneric", bound=AuditEventObject[Any])


class AbstractAuditPublisherService(ABC, Generic[AuditEventGeneric]):
    """Audit publisher service."""

    def __init__(self, sender: ServiceName, publisher: AbstractPublisher[GenericMessage[AuditEventGeneric]]) -> None:
        """Initialize the audit publisher service."""
        self._sender: ServiceName = sender
        self._publisher: AbstractPublisher[GenericMessage[AuditEventGeneric]] = publisher

    @abstractmethod
    def build_routing_key_pattern(self, audit_event: AuditEventGeneric) -> RoutingKey:
        """Build the routing key pattern for the audit event.

        The routing key should follow the pattern: {prefix}.{where}.{what}.{why}
        where:
        - where: The service name from audit_event.where
        - what: The entity name from audit_event.what
        - why: The functional event name from audit_event.why

        Returns:
            RoutingKey: The routing key pattern for routing the message.
        """
        raise NotImplementedError

    async def publish(self, audit_event: AuditEventGeneric) -> None:
        """Publish the audit event."""
        message: GenericMessage[AuditEventGeneric] = GenericMessage(
            data=audit_event,
        )
        routing_key: RoutingKey = self.build_routing_key_pattern(audit_event=audit_event)

        try:
            await self._publisher.publish(message=message, routing_key=routing_key)
        except AiopikaPluginBaseError as exception:
            raise AuditServiceError(
                "Failed to publish the audit event.",
                cause=exception,
                audit_event=audit_event,
                routing_key=routing_key,
            ) from exception



"""Audit service module."""

from .exceptions import AuditServiceError
from .objects import AuditEventObject
from .services import AbstractAuditPublisherService

__all__: list[str] = [
    "AbstractAuditPublisherService",
    "AuditEventObject",
    "AuditServiceError",
]



"""Kratos types module."""

import uuid
from typing import NewType

# Provider is the name of the authentication provider for oidc/saml
KratosIdentityId = NewType("KratosIdentityId", uuid.UUID)
KratosProvider = NewType("KratosProvider", str)
KratosExternalId = NewType("KratosExternalId", str)
KratosSchemaId = NewType("KratosSchemaId", str)
KratosRecoveryCode = NewType("KratosRecoveryCode", str)
KratosRecoveryLink = NewType("KratosRecoveryLink", str)
__all__: list[str] = [
    "KratosExternalId",
    "KratosIdentityId",
    "KratosProvider",
    "KratosSchemaId",
]



"""Provides the Kratos Objects."""

import datetime
import uuid
from typing import ClassVar, Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from .enums import AuthenticationMethodEnum, AuthenticatorAssuranceLevelEnum, KratosIdentityStateEnum
from .types import KratosExternalId, KratosIdentityId, KratosProvider, KratosSchemaId


class KratosTraitsObject(BaseModel):
    """Traits for Kratos.

    Can be extended to include additional traits.

    email: The email address of the user.
    realm_id: The realm ID of the user. (it's the segmentation id for all resources)
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")


class MetadataObject(BaseModel):
    """Metadata for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")


class KratosRecoveryAddressObject(BaseModel):
    """Recovery address for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: uuid.UUID
    value: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    via: str


GenericTraitsObject = TypeVar("GenericTraitsObject", bound=KratosTraitsObject)
GenericMetadataPublicObject = TypeVar("GenericMetadataPublicObject", bound=MetadataObject)
GenericMetadataAdminObject = TypeVar("GenericMetadataAdminObject", bound=MetadataObject)


class KratosIdentityObject(
    BaseModel, Generic[GenericTraitsObject, GenericMetadataPublicObject, GenericMetadataAdminObject]
):
    """Identity for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: KratosIdentityId
    state: KratosIdentityStateEnum
    state_changed_at: datetime.datetime
    traits: GenericTraitsObject
    created_at: datetime.datetime
    updated_at: datetime.datetime
    external_id: KratosExternalId | None = None
    metadata_admin: GenericMetadataAdminObject | None = None
    metadata_public: GenericMetadataPublicObject | None = None
    recovery_addresses: list[KratosRecoveryAddressObject]
    schema_id: KratosSchemaId
    schema_url: str


GenericKratosIdentityObject = TypeVar("GenericKratosIdentityObject", bound=BaseModel)


class KratosAuthenticationMethod(BaseModel):
    """Authentication method for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    aal: AuthenticatorAssuranceLevelEnum
    completed_at: datetime.datetime
    method: AuthenticationMethodEnum
    provider: KratosProvider | None = None


class KratosSessionObject(BaseModel, Generic[GenericKratosIdentityObject]):
    """Session object for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: uuid.UUID
    active: bool
    issued_at: datetime.datetime
    expires_at: datetime.datetime
    authenticated_at: datetime.datetime
    authentication_methods: list[KratosAuthenticationMethod]
    authenticator_assurance_level: AuthenticatorAssuranceLevelEnum
    identity: GenericKratosIdentityObject
    tokenized: str | None = None



"""Provides the core application module for the Python Factory."""

from .application import ApplicationAbstract
from .builder import ApplicationGenericBuilder
from .config import (
    AppCsrfConfig,
    BaseApplicationConfig,
    RootConfig,
)
from .csrf import depends_csrf_protect, register_csrf_protect_exception_handler
from .depends import (
    DependsApplicationConfig,
    DependsRootConfig,
    depends_application_config,
    depends_csrf_config,
    depends_root_config,
)
from .enums import EnvironmentEnum
from .exceptions import ConfigBuilderError, UnableToAcquireApplicationConfigError
from .handlers import register_exception_handlers

__all__: list[str] = [
    "AppCsrfConfig",
    "ApplicationAbstract",
    "ApplicationGenericBuilder",
    "BaseApplicationConfig",
    "ConfigBuilderError",
    "DependsApplicationConfig",
    "DependsRootConfig",
    "EnvironmentEnum",
    "RootConfig",
    "UnableToAcquireApplicationConfigError",
    "depends_application_config",
    "depends_csrf_config",
    "depends_csrf_protect",
    "depends_root_config",
    "register_csrf_protect_exception_handler",
    "register_exception_handlers",
]



"""Provides the JWT bearer token extraction strategies."""

from abc import ABC, abstractmethod

from fastapi import Request

from fastapi_factory_utilities.core.security.types import JWTToken

from .configs import JWTBearerAuthenticationConfig, JWTLocation
from .exceptions import InvalidJWTError, MissingJWTCredentialsError


class JWTBearerTokenExtractionStrategyAbstract(ABC):
    """JWT bearer token extraction strategy."""

    def __init__(self, jwt_bearer_authentication_config: JWTBearerAuthenticationConfig) -> None:
        """Initialize the JWT bearer token extraction strategy."""
        self._jwt_bearer_authentication_config: JWTBearerAuthenticationConfig = jwt_bearer_authentication_config

    @abstractmethod
    def extract_token(self, request: Request) -> JWTToken:
        """Extract the JWT bearer token from the request."""
        raise NotImplementedError()


class JWTBearerTokenExtractionStrategyHeaderAuthorizationBearer(JWTBearerTokenExtractionStrategyAbstract):
    """JWT bearer token extraction strategy for the header authorization bearer."""

    @classmethod
    def extract_authorization_header_from_request(cls, request: Request) -> str:
        """Extract the authorization header from the request.

        Args:
            request (Request): The request object.

        Returns:
            str: The authorization header.

        Raises:
            MissingJWTCredentialsError: If the authorization header is missing.
        """
        authorization_header: str | None = request.headers.get("Authorization", None)
        if not authorization_header:
            raise MissingJWTCredentialsError(message="Missing Credentials")
        return authorization_header

    @classmethod
    def extract_bearer_token_from_authorization_header(cls, authorization_header: str) -> JWTToken:
        """Extract the bearer token from the authorization header.

        Args:
            authorization_header (str): The authorization header.

        Returns:
            JWTToken: The bearer token.

        Raises:
            InvalidJWTError: If the authorization header is invalid.
        """
        if not authorization_header.startswith("Bearer "):
            raise InvalidJWTError(message="Invalid Credentials")
        return JWTToken(authorization_header.split(sep=" ")[1])

    def extract_token(self, request: Request) -> JWTToken:
        """Extract the JWT bearer token from the request."""
        authorization_header: str = self.extract_authorization_header_from_request(request=request)
        return self.extract_bearer_token_from_authorization_header(authorization_header=authorization_header)


class JWTBearerTokenExtractionStrategyCookie(JWTBearerTokenExtractionStrategyAbstract):
    """JWT bearer token extraction strategy for the cookie."""

    def __init__(self, jwt_bearer_authentication_config: JWTBearerAuthenticationConfig) -> None:
        """Initialize the JWT bearer token extraction strategy."""
        super().__init__(jwt_bearer_authentication_config=jwt_bearer_authentication_config)
        if not jwt_bearer_authentication_config.cookie_name:
            raise ValueError("Cookie name is required")
        self._cookie_name: str = jwt_bearer_authentication_config.cookie_name

    def extract_token(self, request: Request) -> JWTToken:
        """Extract the JWT bearer token from the request."""
        if not self._cookie_name:
            raise MissingJWTCredentialsError(message="Missing Credentials")
        cookie: str | None = request.cookies.get(self._cookie_name, None)
        if not cookie:
            raise MissingJWTCredentialsError(message="Missing Credentials")
        return JWTToken(cookie)


class JWTHeaderTokenExtractionStrategy(JWTBearerTokenExtractionStrategyAbstract):
    """JWT header token extraction strategy."""

    def __init__(self, jwt_bearer_authentication_config: JWTBearerAuthenticationConfig) -> None:
        """Initialize the JWT header token extraction strategy."""
        super().__init__(jwt_bearer_authentication_config=jwt_bearer_authentication_config)
        if not jwt_bearer_authentication_config.header_name:
            raise ValueError("Header name is required")
        self._header_name: str = jwt_bearer_authentication_config.header_name

    def extract_token(self, request: Request) -> JWTToken:
        """Extract the JWT bearer token from the request."""
        header_value: str | None = request.headers.get(self._header_name, None)
        if not header_value:
            raise MissingJWTCredentialsError(message="Missing Credentials")
        return JWTToken(header_value)


def get_strategy_from_location(
    jwt_bearer_authentication_config: JWTBearerAuthenticationConfig, location: object
) -> JWTBearerTokenExtractionStrategyAbstract:
    """Get the JWT bearer token extraction strategy from the location."""
    match location:
        case JWTLocation.HEADER:
            return JWTHeaderTokenExtractionStrategy(jwt_bearer_authentication_config=jwt_bearer_authentication_config)
        case JWTLocation.AUTHORIZATION_BEARER:
            return JWTBearerTokenExtractionStrategyHeaderAuthorizationBearer(
                jwt_bearer_authentication_config=jwt_bearer_authentication_config
            )
        case JWTLocation.COOKIE:
            return JWTBearerTokenExtractionStrategyCookie(
                jwt_bearer_authentication_config=jwt_bearer_authentication_config
            )
        case _:
            msg_location: str = str(location)
            raise ValueError(f"Invalid location: {msg_location}")


def extract_token_from_request(
    request: Request, jwt_bearer_authentication_config: JWTBearerAuthenticationConfig
) -> JWTToken:
    """Extract the JWT bearer token from the request."""
    try:
        strategies: list[JWTBearerTokenExtractionStrategyAbstract] = [
            get_strategy_from_location(
                jwt_bearer_authentication_config=jwt_bearer_authentication_config, location=location
            )
            for location in jwt_bearer_authentication_config.authorized_locations
        ]
    except ValueError as error:
        # A ValueError here indicates a misconfigured JWTBearerAuthenticationConfig (e.g., missing header or cookie
        # name for a configured location, or an invalid location value). Surface this as a MissingJWTCredentialsError
        # so that callers like the authentication service can consistently translate it into an HTTP 401 response
        # instead of propagating an unhandled exception.
        raise MissingJWTCredentialsError(message=str(error)) from error
    last_invalid_error: InvalidJWTError | None = None
    for strategy in strategies:
        try:
            return strategy.extract_token(request=request)
        except MissingJWTCredentialsError:
            # Ignore the error and try the next strategy
            continue
        except InvalidJWTError as error:
            # Remember that we saw invalid credentials and try the next strategy
            last_invalid_error = error
            continue
    if last_invalid_error is not None:
        raise last_invalid_error
    raise MissingJWTCredentialsError(message="Missing Credentials")



"""Hydra service module."""

from .exceptions import HydraOperationError, HydraTokenInvalidError
from .objects import HydraTokenIntrospectObject
from .services import (
    HydraIntrospectGenericService,
    HydraOAuth2ClientCredentialsService,
)
from .types import HydraAccessToken, HydraClientId, HydraClientSecret

__all__: list[str] = [
    "HydraAccessToken",
    "HydraClientId",
    "HydraClientSecret",
    "HydraIntrospectGenericService",
    "HydraOAuth2ClientCredentialsService",
    "HydraOperationError",
    "HydraTokenIntrospectObject",
    "HydraTokenInvalidError",
]



"""Provides the Taskiq plugin."""

from collections.abc import Callable
from typing import TYPE_CHECKING

from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract
from fastapi_factory_utilities.core.plugins.taskiq_plugins.exceptions import TaskiqPluginConfigError
from fastapi_factory_utilities.core.utils.redis_configs import (
    RedisCredentialsConfig,
    RedisCredentialsConfigError,
    build_redis_credentials_config,
)

from .depends import DEPENDS_SCHEDULER_COMPONENT_KEY
from .schedulers import SchedulerComponent

if TYPE_CHECKING:
    pass


class TaskiqPlugin(PluginAbstract):
    """Taskiq plugin."""

    def __init__(
        self,
        name_suffix: str,
        redis_credentials_config: RedisCredentialsConfig | None = None,
        register_hook: Callable[[SchedulerComponent], None] | None = None,
    ) -> None:
        """Initialize the Taskiq plugin."""
        super().__init__()
        self._redis_credentials_config: RedisCredentialsConfig | None = redis_credentials_config
        self._register_hook: Callable[[SchedulerComponent], None] | None = register_hook
        self._scheduler_component: SchedulerComponent = SchedulerComponent(name_suffix=name_suffix)

    def on_load(self) -> None:
        """On load."""
        assert self._application is not None
        # Build the Redis credentials configuration if not provided
        if self._redis_credentials_config is None:
            try:
                self._redis_credentials_config = build_redis_credentials_config(application=self._application)
            except RedisCredentialsConfigError as exception:
                raise TaskiqPluginConfigError("Unable to build the Redis credentials configuration.") from exception
        # Configure the scheduler component
        self._scheduler_component.configure(
            redis_connection_string=self._redis_credentials_config.url, app=self._application.get_asgi_app()
        )
        self._add_to_state(key=DEPENDS_SCHEDULER_COMPONENT_KEY, value=self._scheduler_component)
        # Register the hook if provided
        if self._register_hook is not None:
            self._register_hook(self._scheduler_component)

    async def on_startup(self) -> None:
        """On startup."""
        assert self._application is not None
        await self._scheduler_component.startup(app=self._application.get_asgi_app())

    async def on_shutdown(self) -> None:
        """On shutdown."""
        assert self._application is not None
        await self._scheduler_component.shutdown()



"""FastAPI Factory Utilities exceptions."""

import logging
import traceback
from typing import Any

from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv.attributes.exception_attributes import (
    EXCEPTION_MESSAGE,
    EXCEPTION_STACKTRACE,
    EXCEPTION_TYPE,
)
from opentelemetry.trace import Span, get_current_span
from structlog.stdlib import BoundLogger, get_logger


class FastAPIFactoryUtilitiesError(Exception):
    """Base exception for the FastAPI Factory Utilities."""

    FILTERED_ATTRIBUTES: tuple[str, ...] = ()
    DEFAULT_LOGGING_LEVEL: int = logging.ERROR
    DEFAULT_MESSAGE: str | None = None

    @classmethod
    def determine_message(
        cls, default_message: str | None, docstring: str | None, kwargs: dict[str, Any], args: tuple[Any, ...]
    ) -> str:
        """Determine the log message for the exception.

        Args:
            default_message: The default message.
            docstring: The docstring.
            kwargs: The keyword arguments.
            args: The arguments.

        Returns:
            str: The log message.
        """
        # If the message is present in the kwargs, return it (keyword argument)
        if "message" in kwargs and kwargs["message"] is not None:
            return kwargs["message"]

        # If the message is present in the args, return it (positional argument)
        if len(args) > 0 and isinstance(args[0], str):
            return args[0]

        # If the default message is not set, try to extract it from the docstring (class docstring)
        if default_message is None and docstring is not None:
            return docstring.split("\n", maxsplit=1)[0]

        # If the default message is not set, return the default message (class attribute)
        return default_message or "An error occurred"

    @classmethod
    def determine_level(cls, default_level: int, kwargs: dict[str, Any]) -> int:
        """Determine the logging level for the exception.

        Args:
            default_level: The default logging level.
            kwargs: The keyword arguments.

        Returns:
            int: The logging level.
        """
        if "level" in kwargs:
            return kwargs["level"]

        return default_level

    @classmethod
    def determine_safe_attributes(cls, kwargs: dict[str, Any], filtered_attributes: tuple[str, ...]) -> dict[str, Any]:
        """Determine the safe attributes for the exception.

        Args:
            kwargs: The keyword arguments.
            filtered_attributes: The filtered attributes.

        Returns:
            dict[str, Any]: The safe attributes.
        """
        # Always filter out "message" and "level" as they are used internally
        internal_filtered = ("message", "level")
        safe_attributes: dict[str, Any] = {}
        for key, value in kwargs.items():
            if key in filtered_attributes or key in internal_filtered:
                continue
            if not isinstance(value, (str, bool, int, float)):
                safe_attributes[key] = str(value)
            else:
                safe_attributes[key] = value
        return safe_attributes

    def __init__(
        self,
        *args: object,
        **kwargs: Any,
    ) -> None:
        """Instantiate the exception.

        Args:
            *args: The arguments.
            **kwargs: The keyword arguments.

        """
        # Get the logger for the child class's module
        logger: BoundLogger = get_logger(self.__class__.__module__)

        # Extract the message and the level from the kwargs if they are present
        self.message: str = self.determine_message(
            default_message=self.DEFAULT_MESSAGE,
            docstring=self.__doc__,
            kwargs=kwargs,
            args=args,
        )
        self.level: int = self.determine_level(
            default_level=self.DEFAULT_LOGGING_LEVEL,
            kwargs=kwargs,
        )

        # Filter the kwargs to remove the filtered attributes
        safe_attributes: dict[str, Any] = self.determine_safe_attributes(
            kwargs=kwargs,
            filtered_attributes=self.FILTERED_ATTRIBUTES,
        )

        # Set the kwargs as attributes of the exception
        for key, value in safe_attributes.items():
            setattr(self, key, value)

        # Log the Exception using the logger for the child class's module
        if self.message:
            logger.log(level=self.level, event=self.message, **safe_attributes)

        try:
            # Propagate the exception
            span: Span = get_current_span()
            # If not otel is setup, INVALID_SPAN is retrieved from get_current_span
            # and it will respond False to the is_recording method
            if span.is_recording():
                # Set the kwargs attributes
                for key, value in safe_attributes.items():
                    span.set_attribute(key, value)

                # Record official Attributes last to avoid overriding them
                span.record_exception(self)
                # Set the exception and error attributes
                span.set_attribute(ERROR_TYPE, self.__class__.__name__)
                span.set_attribute(EXCEPTION_MESSAGE, self.message)
                span.set_attribute(EXCEPTION_STACKTRACE, traceback.format_exc())
                span.set_attribute(EXCEPTION_TYPE, self.__class__.__name__)
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Suppress any errors that occur while propagating the exception
            logger.error("An error occurred while recording the exception as trace", exc_info=e)

        # Call the parent class with the message so str(exception) returns it
        super().__init__(self.message)



"""Provides utility functions for query parameters helpers.

Objectives:
- Provide a way to simplify the process to code a search endpoint with query parameters filtering.
"""

from typing import Any

from fastapi import Request


class QueryFilterValidationError(ValueError):
    """Exception raised when a query filter is invalid."""


class QueryFilterUnauthorizedError(ValueError):
    """Exception raised when a query filter is unauthorized."""


class QueryFilterHelper:
    """Helper class to filter query parameters as a Dependency."""

    def __init__(
        self,
        authorized_filters: dict[str, type],
        raise_on_unauthorized_filter: bool = True,
        raise_on_invalid_filter: bool = True,
    ) -> None:
        """Initialize the QueryFilterHelper.

        Args:
            authorized_filters (dict[str, type]): The authorized filters.
            raise_on_unauthorized_filter (bool): Whether to raise an exception if an unauthorized filter is provided.
            raise_on_invalid_filter (bool): Whether to raise an exception if an invalid filter is provided.
        """
        self._authorized_filters: dict[str, type] = authorized_filters
        self._raise_on_unauthorized_filter: bool = raise_on_unauthorized_filter
        self._raise_on_invalid_filter: bool = raise_on_invalid_filter
        self._filters: dict[str, Any] = {}

    def _raise_on_unauthorized_filter_error(self, key: str) -> None:
        """Raise an unauthorized filter exception.

        Args:
            key (str): The key of the unauthorized filter.
        """
        if self._raise_on_unauthorized_filter:
            raise QueryFilterUnauthorizedError(f"Unauthorized filter: {key}")

    def _raise_on_invalid_filter_error(self, key: str, value: Any, error: Exception) -> None:
        """Raise an invalid filter exception.

        Args:
            key (str): The key of the invalid filter.
            value (Any): The value of the invalid filter.
            error (Exception): The error that occurred while transforming the filter.

        Raises:
            QueryFilterValidationError: If the filter is invalid and raise_on_invalid_filter is True.
        """
        if self._raise_on_invalid_filter:
            raise QueryFilterValidationError(f"Invalid filter: {key} with value: {value}") from error

    def _transform_filter(self, key: str, value: Any, filter_type: type) -> Any | None:
        """Transform the filter.

        Args:
            key (str): The key of the filter.
            value (Any): The value of the filter.
            filter_type (type): The type of the filter.

        Returns:
            Any: The transformed filter.

        Raises:
            QueryFilterValidationError: If the filter is invalid.
        """
        # If the value is already of the correct type, return it
        if isinstance(value, filter_type):
            return value
        try:
            return filter_type(value)
        except ValueError as e:
            self._raise_on_invalid_filter_error(key=key, value=value, error=e)
        return None

    def validate_filters(self, filters: dict[str, Any]) -> dict[str, Any]:
        """Validate the filters.

        Args:
            filters (dict[str, Any]): The filters.

        Returns:
            dict[str, Any]: The validated filters.

        Raises:
            QueryFilterUnauthorizedError: If an unauthorized filter is provided
            and raise_on_unauthorized_filter is True.
            QueryFilterValidationError: If an invalid filter is provided
            and raise_on_invalid_filter is True.
        """
        validated_filters: dict[str, Any] = {}
        for key, value in filters.items():
            if key not in self._authorized_filters:
                self._raise_on_unauthorized_filter_error(key=key)
            transformed_value: Any | None = self._transform_filter(
                key=key, value=value, filter_type=self._authorized_filters[key]
            )
            if transformed_value is not None:
                validated_filters[key] = transformed_value
        return validated_filters

    def __call__(self, request: Request) -> dict[str, Any]:
        """Call the QueryFilterHelper."""
        self._filters = self.validate_filters(filters=dict(request.query_params.items()))
        return self._filters


__all__: list[str] = [
    "QueryFilterHelper",
    "QueryFilterUnauthorizedError",
    "QueryFilterValidationError",
]



"""Provides the dependencies for the Aiohttp plugin."""

from fastapi import Request
from fastapi.datastructures import State

from .constants import STATE_PREFIX_KEY
from .exceptions import AioHttpClientResourceNotFoundError
from .resources import AioHttpClientResource


class AioHttpResourceDepends:
    """Aiohttp client depends."""

    def __init__(self, key: str) -> None:
        """Initialize the Aiohttp client depends."""
        self._key: str = key

    @classmethod
    def export_from_state(cls, state: State, key: str) -> AioHttpClientResource:
        """Export the Aiohttp resource from the state."""
        resource: AioHttpClientResource | None = getattr(state, f"{STATE_PREFIX_KEY}{key}", None)
        if resource is None:
            raise AioHttpClientResourceNotFoundError("Aiohttp resource not found in the application state.", key=key)
        return resource

    def __call__(self, request: Request) -> AioHttpClientResource:
        """Get the Aiohttp resource."""
        resource: AioHttpClientResource = self.export_from_state(state=request.app.state, key=self._key)
        return resource



"""Provides the KratosService class for handling Kratos operations."""

import datetime
import json
from http import HTTPStatus
from typing import Any, ClassVar, Generic, TypeVar, get_args

import aiohttp
from pydantic import BaseModel, ConfigDict, Field, ValidationError
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins.aiohttp import AioHttpClientResource
from fastapi_factory_utilities.core.utils.ory import get_next_page_token_from_link_header

from .enums import AuthenticationMethodEnum, KratosFlowTypeEnum, KratosIdentityPatchOpEnum
from .exceptions import KratosOperationError, KratosSessionInvalidError
from .types import KratosIdentityId, KratosRecoveryCode, KratosRecoveryLink

_logger: BoundLogger = get_logger(__package__)

GenericKratosSessionObject = TypeVar("GenericKratosSessionObject", bound=BaseModel)


class KratosGenericWhoamiService(Generic[GenericKratosSessionObject]):
    """Service class for handling Kratos operations."""

    COOKIE_NAME: str = "ory_kratos_session"

    def __init__(self, kratos_public_http_resource: AioHttpClientResource) -> None:
        """Initialize the KratosService class.

        Args:
            kratos_public_http_resource (AioHttpClientResource): Kratos public HTTP resource.
        """
        self._kratos_public_http_resource: AioHttpClientResource = kratos_public_http_resource
        # Retrieve the concrete introspect object class
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._concreate_session_object_class: type[GenericKratosSessionObject] = generic_args[0]

    async def whoami(self, cookie_value: str) -> GenericKratosSessionObject:
        """Get the current user session.

        Args:
            cookie_value (str): Cookie value.

        Returns:
            GenericKratosSessionObject: Kratos session object.

        Raises:
            KratosOperationError: If the Kratos service returns an error.
            KratosSessionInvalidError: If the Kratos session is invalid.
        """
        cookies: dict[str, str] = {self.COOKIE_NAME: cookie_value}
        async with self._kratos_public_http_resource.acquire_client_session(cookies=cookies) as session:
            async with session.get(
                url="/sessions/whoami",
            ) as response:
                if response.status >= HTTPStatus.INTERNAL_SERVER_ERROR.value:
                    raise KratosOperationError(message=f"Kratos service error: {response.status} - {response.reason}")
                if response.status == HTTPStatus.UNAUTHORIZED:
                    raise KratosSessionInvalidError(
                        message=f"Kratos session invalid: {response.status} - {response.reason}"
                    )
                if response.status != HTTPStatus.OK:
                    raise KratosOperationError(message=f"Kratos service error: {response.status} - {response.reason}")

                try:
                    kratos_session: GenericKratosSessionObject = self._concreate_session_object_class.model_validate(
                        await response.json()
                    )
                except ValidationError as e:
                    raise KratosOperationError(message=f"Kratos service error: {e}") from e

                return kratos_session


GenericKratosIdentityObject = TypeVar("GenericKratosIdentityObject", bound=BaseModel)


class KratosIdentityPatchObject(BaseModel):
    """Patch object for Kratos identity."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid", populate_by_name=True)

    from_: str | None = Field(alias="from", default=None)
    op: KratosIdentityPatchOpEnum
    path: str
    value: Any | None = None


class KratosIdentityGenericService(Generic[GenericKratosIdentityObject, GenericKratosSessionObject]):
    """Service class for handling Kratos identity operations."""

    IDENTITY_ENDPOINT: str = "/admin/identities"
    ADMIN_ENDPOINT: str = "/admin"

    def __init__(self, kratos_admin_http_resource: AioHttpClientResource) -> None:
        """Initialize the KratosIdentityGenericService class.

        Args:
            kratos_admin_http_resource (AioHttpClientResource): Kratos admin HTTP resource.
        """
        self._kratos_admin_http_resource: AioHttpClientResource = kratos_admin_http_resource
        # Retrieve the concrete introspect object class
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._concreate_identity_object_class: type[GenericKratosIdentityObject] = generic_args[0]
        self._concreate_session_object_class: type[GenericKratosSessionObject] = generic_args[1]

    async def get_identity(self, identity_id: KratosIdentityId) -> GenericKratosIdentityObject:
        """Get a Kratos identity by ID.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to get.

        Returns:
            GenericKratosIdentityObject: Kratos identity object.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.get(url=f"{self.IDENTITY_ENDPOINT}/{identity_id}") as response:
                    response.raise_for_status()
                    return self._concreate_identity_object_class.model_validate(await response.json())
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to get the Kratos identity",
                identity_id=identity_id,
                status_code=status_code,
            ) from e

    async def create_identity(self, identity: GenericKratosIdentityObject) -> GenericKratosIdentityObject:
        """Create a Kratos identity."""
        raise NotImplementedError("Not implemented")

    async def update_identity(
        self, identity_id: KratosIdentityId, patches: list[KratosIdentityPatchObject]
    ) -> GenericKratosIdentityObject:
        """Update a Kratos identity.

        Raises:
            KratosOperationError: If the Kratos service returns an error.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to update.
            patches (list[KratosIdentityPatchObject]): The patches to apply to the identity.

        Returns:
            GenericKratosIdentityObject: The updated identity.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                patches_dict: list[dict[str, Any]] = [
                    patch.model_dump(by_alias=True, exclude_none=True) for patch in patches
                ]
                async with session.patch(url=f"{self.IDENTITY_ENDPOINT}/{identity_id}", json=patches_dict) as response:
                    response.raise_for_status()
                    identity: GenericKratosIdentityObject = self._concreate_identity_object_class.model_validate(
                        await response.json()
                    )
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to update the Kratos identity",
                identity_id=identity_id,
                patches=patches,
                status_code=status_code,
            ) from e

        _logger.info(
            "Identity updated successfully",
            identity_id=identity_id,
            patches=patches,
            identity=identity,
        )

        return identity

    async def delete_identity_credentials(
        self,
        identity_id: KratosIdentityId,
        credentials_type: AuthenticationMethodEnum,
        identifier: str | None = None,
    ) -> None:
        """Delete the credentials of a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to delete the credentials of.
            credentials_type (AuthenticationMethodEnum): The type of credentials to delete.
            identifier (str | None): The identifier of the credentials to delete.
            (Mandatory for OIDC and SAML credentials)

        Returns:
            None: If the credentials are deleted successfully.
        """
        if credentials_type not in [AuthenticationMethodEnum.OIDC, AuthenticationMethodEnum.SAML]:
            if identifier is not None:
                raise ValueError("Identifier is only supported for OIDC and SAML credentials")
        elif identifier is None:
            raise ValueError("Identifier is mandatory for OIDC and SAML credentials")

        query_params: dict[str, str] = {}
        if identifier is not None:
            query_params["identifier"] = identifier

        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.delete(
                    url=f"{self.IDENTITY_ENDPOINT}/{identity_id}/credentials", params=query_params
                ) as response:
                    response.raise_for_status()
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to delete the credentials of the Kratos identity",
                identity_id=identity_id,
                credentials_type=credentials_type,
                identifier=identifier,
                status_code=status_code,
            ) from e

        _logger.info(
            "Credentials deleted successfully",
            identity_id=identity_id,
            credentials_type=credentials_type,
            identifier=identifier,
        )

    async def delete_identity_sessions(self, identity_id: KratosIdentityId) -> None:
        """Delete the sessions of a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to delete the sessions of.

        Returns:
            None: If the sessions are deleted successfully.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.delete(url=f"{self.IDENTITY_ENDPOINT}/{identity_id}/sessions") as response:
                    response.raise_for_status()
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to delete the sessions of the Kratos identity",
                identity_id=identity_id,
                status_code=status_code,
            ) from e

        _logger.info(
            "Sessions deleted successfully",
            identity_id=identity_id,
        )

    async def delete_identity(self, identity_id: KratosIdentityId) -> None:
        """Delete a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to delete.

        Returns:
            None: If the identity is deleted successfully.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.delete(url=f"{self.IDENTITY_ENDPOINT}/{identity_id}") as response:
                    response.raise_for_status()
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to delete the Kratos identity",
                identity_id=identity_id,
                status_code=status_code,
            ) from e

        _logger.info("Identity deleted successfully", identity_id=identity_id)

    async def list_sessions(
        self,
        identity_id: KratosIdentityId,
        active: bool = True,
        page_size: int = 250,
        page_token: str | None = None,
    ) -> tuple[list[GenericKratosSessionObject], str | None]:
        """List the sessions of a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to list the sessions of.
            active (bool): Whether to filter for active sessions only. Defaults to True.
            page_size (int): The number of sessions to return per page. Defaults to 250.
            page_token (str | None): The page token to use for pagination.

        Returns:
            Tuple[list[GenericKratosSessionObject], str | None]: A tuple containing the list of
                sessions and the next page token.
        """
        query_params: dict[str, str] = {
            "page_size": str(page_size),
            "active": str(active),
        }
        if page_token is not None:
            query_params["page_token"] = page_token

        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.get(
                    url=f"{self.IDENTITY_ENDPOINT}/{identity_id}/sessions", params=query_params
                ) as response:
                    response.raise_for_status()
                    sessions: list[GenericKratosSessionObject] = [
                        self._concreate_session_object_class.model_validate(session)
                        for session in await response.json()
                    ]
                    link_header: str | None = response.headers.get("Link", None)
                    next_page_token: str | None = get_next_page_token_from_link_header(link_header=link_header)

                    return sessions, next_page_token
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to list the sessions of the Kratos identity",
                identity_id=identity_id,
                status_code=status_code,
            ) from e

    async def create_recovery_code(
        self,
        identity_id: KratosIdentityId,
        flow_type: KratosFlowTypeEnum,
        expires_in: datetime.timedelta,
    ) -> KratosRecoveryCode:
        """Create a recovery code for a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to create the recovery code for.
            flow_type (KratosFlowTypeEnum): The type of flow to create the recovery code for.
            expires_in (datetime.timedelta): The expiration time for the recovery code.

        Raises:
            KratosOperationError: If the Kratos service returns an error.

        Returns:
            KratosRecoveryCode: The recovery code.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.post(
                    url=f"{self.ADMIN_ENDPOINT}/recovery/code",
                    json={
                        "flow_type": flow_type.value,
                        "expires_in": f"{int(expires_in.total_seconds())}s",
                        "identity_id": str(identity_id),
                    },
                ) as response:
                    response.raise_for_status()
                    code: KratosRecoveryCode = KratosRecoveryCode(await response.json())
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to create the recovery code for the Kratos identity",
                identity_id=identity_id,
                flow_type=flow_type,
                expires_in=expires_in,
                status_code=status_code,
            ) from e

        _logger.info(
            "Recovery code created successfully",
            identity_id=identity_id,
            flow_type=flow_type,
            expires_in=expires_in,
        )

        return code

    async def create_recovery_link(
        self,
        identity_id: KratosIdentityId,
        expires_in: datetime.timedelta,
    ) -> KratosRecoveryLink:
        """Create a recovery link for a Kratos identity.

        Args:
            identity_id (KratosIdentityId): The ID of the identity to create the recovery link for.
            expires_in (datetime.timedelta): The expiration time for the recovery link.

        Returns:
            KratosRecoveryLink: The recovery link.

        Raises:
            KratosOperationError: If the Kratos service returns an error.
        """
        try:
            async with self._kratos_admin_http_resource.acquire_client_session() as session:
                async with session.post(
                    url=f"{self.ADMIN_ENDPOINT}/recovery/link",
                    json={
                        "expires_in": f"{int(expires_in.total_seconds())}s",
                        "identity_id": str(identity_id),
                    },
                ) as response:
                    response.raise_for_status()
                    link: KratosRecoveryLink = KratosRecoveryLink(await response.json())
        except (aiohttp.ClientResponseError, json.JSONDecodeError, ValidationError) as e:
            status_code: str | None = getattr(e, "status", None)
            raise KratosOperationError(
                message="Failed to create the recovery link for the Kratos identity",
                identity_id=identity_id,
                expires_in=expires_in,
                status_code=status_code,
            ) from e

        _logger.info(
            "Recovery link created successfully",
            identity_id=identity_id,
            expires_in=expires_in,
        )

        return link



"""Provides the application handlers."""

from http import HTTPStatus
from typing import cast

from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from structlog.stdlib import BoundLogger, get_logger

_logger: BoundLogger = get_logger(__package__)


async def validation_exception_handler(request: Request, exc: Exception) -> JSONResponse:
    """Validation exception handler."""
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    content = {"status_code": 422, "message": exc_str, "data": None}
    errors = cast(RequestValidationError, exc).errors()
    _logger.error("Validation error", request=request, exc_str=exc_str, content=content)
    return JSONResponse(status_code=HTTPStatus.BAD_REQUEST, content={"detail": errors})


def register_exception_handlers(app: FastAPI) -> None:
    """Register the exception handlers."""
    app.add_exception_handler(RequestValidationError, validation_exception_handler)



"""Aiopika Plugin Module."""

from .depends import depends_aiopika_robust_connection
from .exceptions import AiopikaPluginBaseError, AiopikaPluginConfigError
from .exchange import Exchange
from .listener import AbstractListener
from .message import GenericMessage
from .plugins import AiopikaPlugin
from .publisher import AbstractPublisher
from .queue import Queue
from .types import ExchangeName, QueueName, RoutingKey

__all__: list[str] = [
    "AbstractListener",
    "AbstractPublisher",
    "AiopikaPlugin",
    "AiopikaPluginBaseError",
    "AiopikaPluginConfigError",
    "Exchange",
    "ExchangeName",
    "GenericMessage",
    "Queue",
    "QueueName",
    "RoutingKey",
    "depends_aiopika_robust_connection",
]



"""Provides the exceptions for the Taskiq plugin."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class TaskiqPluginBaseError(FastAPIFactoryUtilitiesError):
    """Base class for all exceptions raised by the Taskiq plugin."""


class TaskiqPluginConfigError(TaskiqPluginBaseError):
    """Exception for the Taskiq plugin configuration."""



"""Provides the Books API."""

from http import HTTPStatus
from typing import cast
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Response

from fastapi_factory_utilities.example.entities.books import BookEntity
from fastapi_factory_utilities.example.services.books import (
    BookService,
    depends_book_service,
)

from .requests import BookCreateRequest, BookUpdateRequest
from .responses import BookListReponse, BookResponseModel

api_v1_books_router: APIRouter = APIRouter(prefix="/books")
api_v2_books_router: APIRouter = APIRouter(prefix="/books")


@api_v1_books_router.get(path="", response_model=BookListReponse)
async def get_books(
    books_service: BookService = Depends(depends_book_service),
) -> BookListReponse:
    """Get all books.

    Args:
        books_service: Book service.

    Returns:
        List of books.
    """
    books: list[BookEntity] = await books_service.get_all_books()

    return BookListReponse(
        books=cast(
            list[BookResponseModel],
            map(lambda book: BookResponseModel(**book.model_dump()), books),
        ),
        size=len(books),
    )


@api_v1_books_router.get(path="/{book_id}", response_model=BookResponseModel)
async def get_book(
    book_id: UUID,
    books_service: BookService = Depends(depends_book_service),
) -> BookResponseModel:
    """Get a book.

    Args:
        book_id: Book id.
        books_service: Book service.

    Returns:
        Book.

    Raises:
        HTTPException: 404 if book not found.
    """
    try:
        book: BookEntity = await books_service.get_book(book_id)
        return BookResponseModel(**book.model_dump())
    except ValueError as e:
        raise HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e)) from e


@api_v1_books_router.post(path="", response_model=BookResponseModel, status_code=HTTPStatus.CREATED)
async def create_book(
    request: BookCreateRequest,
    books_service: BookService = Depends(depends_book_service),
) -> BookResponseModel:
    """Create a new book.

    Args:
        request: Book creation request.
        books_service: Book service.

    Returns:
        Created book.
    """
    book_entity: BookEntity = BookEntity(
        title=request.title,
        book_type=request.book_type,
    )

    created_book: BookEntity = await books_service.add_book(book=book_entity)

    return BookResponseModel(**created_book.model_dump())


@api_v1_books_router.put(path="/{book_id}", response_model=BookResponseModel)
async def update_book(
    book_id: UUID,
    request: BookUpdateRequest,
    books_service: BookService = Depends(depends_book_service),
) -> BookResponseModel:
    """Update an existing book.

    Args:
        book_id: Book id.
        request: Book update request.
        books_service: Book service.

    Returns:
        Updated book.

    Raises:
        HTTPException: 404 if book not found.
    """
    try:
        book_entity: BookEntity = await books_service.get_book(book_id=book_id)
        updated_book_entity: BookEntity = book_entity.model_copy(
            update={"title": request.title, "book_type": request.book_type}
        )
        updated_book: BookEntity = await books_service.update_book(book=updated_book_entity)

        return BookResponseModel(**updated_book.model_dump())
    except ValueError as e:
        raise HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e)) from e


@api_v1_books_router.delete(path="/{book_id}", status_code=HTTPStatus.NO_CONTENT)
async def delete_book(
    book_id: UUID,
    books_service: BookService = Depends(depends_book_service),
) -> Response:
    """Delete a book.

    Args:
        book_id: Book id.
        books_service: Book service.

    Returns:
        No content response.

    Raises:
        HTTPException: 404 if book not found.
    """
    try:
        await books_service.remove_book(book_id=book_id)
        return Response(status_code=HTTPStatus.NO_CONTENT)
    except ValueError as e:
        raise HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e)) from e



"""Provides the JWT bearer token objects."""

import datetime
from typing import Annotated, Any, ClassVar

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from fastapi_factory_utilities.core.security.types import (
    OAuth2Audience,
    OAuth2Issuer,
    OAuth2Scope,
    OAuth2Subject,
)


def validate_string_list_field(value: Any) -> list[str]:
    """Validate a string list field.

    Accepts either a space-separated string or a list of strings.
    Converts all values to lowercase strings.

    Args:
        value: Either a string (space-separated) or a list of strings.

    Returns:
        A list of lowercase strings.

    Raises:
        ValueError: If the value is not a string or list, or if the resulting list is empty.
    """
    cleaned_value: list[str]
    if isinstance(value, str):
        cleaned_value = value.split(sep=" ")
    elif isinstance(value, list):
        cleaned_value = [str(item) for item in value if item is not None]
    else:
        raise ValueError(f"Invalid value type: expected str or list, got {type(value).__name__}")
    cleaned_value = [item.lower() for item in cleaned_value if item.strip()]
    if len(cleaned_value) == 0:
        raise ValueError("Invalid value: empty list after processing")
    return cleaned_value


def validate_timestamp_field(value: Any) -> datetime.datetime:
    """Validate a timestamp field.

    Accepts either a Unix timestamp (int or string) or a datetime object.
    Converts timestamps to UTC datetime objects.

    Args:
        value: Either a Unix timestamp (int or string) or a datetime object.

    Returns:
        A datetime object in UTC timezone.

    Raises:
        ValueError: If the value cannot be converted to a datetime.
    """
    if isinstance(value, datetime.datetime):
        return value
    if isinstance(value, str):
        try:
            value = int(value)
        except ValueError as e:
            raise ValueError(f"Invalid timestamp string: {value}") from e
    if isinstance(value, int):
        try:
            return datetime.datetime.fromtimestamp(value, tz=datetime.UTC)
        except (ValueError, OSError) as e:
            raise ValueError(f"Invalid timestamp value: {value}") from e
    raise ValueError(f"Invalid value type: expected int, str, or datetime, got {type(value).__name__}")


class JWTPayload(BaseModel):
    """JWT bearer token payload.

    Represents a decoded JWT bearer token with OAuth2 claims.
    All fields are required and validated according to OAuth2/JWT standards.

    Attributes:
        scope: List of OAuth2 scopes granted by the token.
        aud: List of audiences (intended recipients) of the token.
        iss: The issuer of the JWT token.
        exp: The expiration date/time of the JWT token (UTC).
        iat: The issued at date/time of the JWT token (UTC).
        nbf: The not before date/time of the JWT token (UTC).
        sub: The subject (user identifier) of the JWT token.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        arbitrary_types_allowed=True,
        extra="ignore",
        frozen=True,
    )

    scp: Annotated[list[OAuth2Scope], BeforeValidator(validate_string_list_field)] = Field(
        description="The scope of the JWT token."
    )
    aud: Annotated[list[OAuth2Audience], BeforeValidator(validate_string_list_field)] = Field(
        description="The audiences of the JWT token."
    )
    iss: OAuth2Issuer = Field(description="The issuer of the JWT token.")
    exp: Annotated[datetime.datetime, BeforeValidator(validate_timestamp_field)] = Field(
        description="The expiration date of the JWT token."
    )
    iat: Annotated[datetime.datetime, BeforeValidator(validate_timestamp_field)] = Field(
        description="The issued at date of the JWT token."
    )
    nbf: Annotated[datetime.datetime, BeforeValidator(validate_timestamp_field)] = Field(
        description="The not before date of the JWT token."
    )
    sub: OAuth2Subject = Field(description="The subject of the JWT token.")



"""Provides a service to interact with the Hydra service."""

import json
from base64 import b64encode
from typing import Any, Generic, TypeVar, get_args

import aiohttp
import jwt
from pydantic import ValidationError

from fastapi_factory_utilities.core.plugins.aiohttp import (
    AioHttpClientResource,
)
from fastapi_factory_utilities.core.security.jwt import JWTBearerAuthenticationConfig
from fastapi_factory_utilities.core.security.types import OAuth2Issuer

from .exceptions import HydraOperationError
from .objects import HydraTokenIntrospectObject
from .types import HydraAccessToken, HydraClientId, HydraClientSecret

HydraIntrospectObjectGeneric = TypeVar("HydraIntrospectObjectGeneric", bound=HydraTokenIntrospectObject)


class HydraIntrospectGenericService(Generic[HydraIntrospectObjectGeneric]):
    """Service to interact with the Hydra introspect service with a generic introspect object."""

    INTROSPECT_ENDPOINT: str = "/admin/oauth2/introspect"
    WELLKNOWN_JWKS_ENDPOINT: str = "/.well-known/jwks.json"

    def __init__(
        self,
        identifier: str,
        config: JWTBearerAuthenticationConfig,
        hydra_admin_http_resource: AioHttpClientResource,
        hydra_public_http_resource: AioHttpClientResource,
    ) -> None:
        """Instanciate the Hydra introspect service.

        Args:
            identifier (str): The identifier of the Hydra introspect service.
            config (JWTBearerAuthenticationConfig): The JWT bearer authentication configuration.
            hydra_admin_http_resource (AioHttpClientResource): The Hydra admin HTTP resource.
            hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.
        """
        self._identifier: str = identifier
        self._config: JWTBearerAuthenticationConfig = config
        self._hydra_admin_http_resource: AioHttpClientResource = hydra_admin_http_resource
        self._hydra_public_http_resource: AioHttpClientResource = hydra_public_http_resource
        # Retrieve the concrete introspect object class
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._concreate_introspect_object_class: type[HydraIntrospectObjectGeneric] = generic_args[0]

    async def introspect(self, token: HydraAccessToken) -> HydraIntrospectObjectGeneric:
        """Introspects a token using the Hydra introspect service.

        Args:
            token (str): The token to introspect.
        """
        try:
            async with self._hydra_admin_http_resource.acquire_client_session() as session:
                async with session.post(
                    url=self.INTROSPECT_ENDPOINT,
                    data={"token": token},
                ) as response:
                    response.raise_for_status()
                    instrospect: HydraIntrospectObjectGeneric = self._concreate_introspect_object_class.model_validate(
                        await response.json()
                    )
        except aiohttp.ClientResponseError as error:
            raise HydraOperationError(
                "An error occurred while introspecting the token", status_code=error.status
            ) from error
        except json.JSONDecodeError as error:
            raise HydraOperationError("An error occurred while decoding the introspect response") from error
        except ValidationError as error:
            raise HydraOperationError("An error occurred while validating the introspect response") from error
        except Exception as error:
            raise HydraOperationError("An error occurred while introspecting the token") from error

        return instrospect

    def get_issuer(self) -> OAuth2Issuer:
        """Get the issuer from the JWT bearer authentication configuration."""
        return self._config.issuer

    async def get_wellknown_jwks(self) -> list[jwt.PyJWK]:
        """Get the JWKS from the Hydra service."""
        try:
            async with self._hydra_public_http_resource.acquire_client_session() as session:
                async with session.get(
                    url=self.WELLKNOWN_JWKS_ENDPOINT,
                ) as response:
                    response.raise_for_status()
                    jwks_data: dict[str, Any] = await response.json()
                    jwks: jwt.PyJWKSet = jwt.PyJWKSet.from_dict(jwks_data)
                    return jwks.keys  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
        except aiohttp.ClientResponseError as error:
            raise HydraOperationError(
                "Failed to get the JWKS from the Hydra service", status_code=error.status
            ) from error
        except json.JSONDecodeError as error:
            raise HydraOperationError("Failed to decode the JWKS from the Hydra service") from error
        except ValidationError as error:
            raise HydraOperationError("Failed to validate the JWKS from the Hydra service") from error
        except Exception as error:
            raise HydraOperationError("Failed to get the JWKS from the Hydra service") from error


class HydraIntrospectService(HydraIntrospectGenericService[HydraTokenIntrospectObject]):
    """Service to interact with the Hydra introspect service with the default HydraTokenIntrospectObject."""


class HydraOAuth2ClientCredentialsService:
    """Service to interact with the Hydra service."""

    INTROSPECT_ENDPOINT: str = "/admin/oauth2/introspect"
    CLIENT_CREDENTIALS_ENDPOINT: str = "/oauth2/token"

    def __init__(
        self,
        identifier: str,
        hydra_public_http_resource: AioHttpClientResource,
        config: JWTBearerAuthenticationConfig,
        default_audience: str,
    ) -> None:
        """Instanciate the Hydra service.

        Args:
            identifier (str): The identifier of the Hydra OAuth2 client credentials service.
            hydra_public_http_resource (AioHttpClientResource): The Hydra public HTTP resource.
            config (JWTBearerAuthenticationConfig): The JWT bearer authentication configuration.
            default_audience (str): The default audience.
        """
        self._identifier: str = identifier
        self._hydra_public_http_resource: AioHttpClientResource = hydra_public_http_resource
        self._config: JWTBearerAuthenticationConfig = config
        self._default_audience: str = default_audience

    @classmethod
    def build_bearer_header(cls, client_id: HydraClientId, client_secret: HydraClientSecret) -> str:
        """Build the bearer header.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.

        Returns:
            str: The bearer header.
        """
        auth_string: str = f"{client_id}:{client_secret}"
        auth_bytes: bytes = auth_string.encode("utf-8")
        auth_b64: str = b64encode(auth_bytes).decode("utf-8")
        return f"Basic {auth_b64}"

    async def oauth2_client_credentials(
        self, client_id: HydraClientId, client_secret: HydraClientSecret, scopes: list[str], audience: str | None = None
    ) -> HydraAccessToken:
        """Get the OAuth2 client credentials.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.
            scopes (list[str]): The scopes.
            audience (str, optional): The audience. Defaults to None.

        Returns:
            str: The access token.

        Raises:
            HydraOperationError: If the client credentials request fails.
        """
        enforced_audience: str = audience if audience is not None else self._default_audience
        try:
            async with self._hydra_public_http_resource.acquire_client_session() as session:
                async with session.post(
                    url=self.CLIENT_CREDENTIALS_ENDPOINT,
                    headers={"Authorization": self.build_bearer_header(client_id, client_secret)},
                    data={
                        "grant_type": "client_credentials",
                        "scope": " ".join(scopes),
                        "audience": enforced_audience,
                    },
                ) as response:
                    response.raise_for_status()
                    response_data = await response.json()
                    return response_data["access_token"]
        except aiohttp.ClientResponseError as error:
            raise HydraOperationError(
                "An error occurred while getting the client credentials", status_code=error.status
            ) from error
        except json.JSONDecodeError as error:
            raise HydraOperationError(
                "An error occurred while getting the client credentials, invalid JSON response"
            ) from error
        except ValidationError as error:
            raise HydraOperationError(
                "An error occurred while getting the client credentials, invalid response"
            ) from error
        except Exception as error:
            raise HydraOperationError(
                "An error occurred while getting the client credentials, unknown error"
            ) from error



"""Mocker Utilities for the Aiohttp plugin."""

import types
from collections.abc import Callable
from http import HTTPStatus
from http.cookies import SimpleCookie
from typing import Any
from unittest.mock import AsyncMock, MagicMock

import aiohttp

from fastapi_factory_utilities.core.plugins.aiohttp.resources import AioHttpClientResource


def build_mocked_aiohttp_response(  # noqa: PLR0913
    status: HTTPStatus,
    json: dict[str, Any] | list[Any] | str | None = None,
    text: str | None = None,
    headers: dict[str, str] | None = None,
    read: bytes | None = None,
    cookies: dict[str, str] | None = None,
    error_message: str | None = None,
    reason: str | None = None,
    url: str | None = None,
    method: str | None = None,
) -> aiohttp.ClientResponse:
    """Build the mocked Aiohttp response.

    Creates a mock aiohttp.ClientResponse object with configurable status code,
    response body (JSON, text, or binary content), headers, and cookies. The mock will
    raise ClientResponseError when raise_for_status() is called if the status
    code indicates an error (>= 400).

    The response supports both usage patterns:
    - Direct await: `response = await session.get(url)`
    - Context manager: `async with session.get(url) as response:`

    Args:
        status: HTTP status code for the response.
        json: JSON data to return from response.json(). Can be a dict, list, or string.
        text: Text data to return from response.text().
        headers: HTTP headers to return from response.headers. Note: In context manager
            pattern, headers is accessed as a property, not a method.
        read: Binary content to return from response.read(). This is the proper aiohttp
            method for reading binary content.
        cookies: Cookies to set in the response. Can be a dict of cookie name-value pairs.
            The cookies will be accessible via response.cookies property.
        error_message: Custom error message for ClientResponseError. If None,
            defaults to a descriptive message based on the status code.
        reason: HTTP reason phrase (e.g., "OK", "Not Found"). If None,
            defaults to the standard phrase for the status code.
        url: The URL of the response (e.g., the final URL after redirects).
        method: The HTTP method used for the request (e.g., "GET", "POST").

    Returns:
        aiohttp.ClientResponse: A mocked ClientResponse object configured with
            the specified parameters. Supports async context manager protocol.

    Example:
        Direct await pattern:
        >>> response = build_mocked_aiohttp_response(status=HTTPStatus.OK, json={"message": "Success"})
        >>> # response = await session.get(url)
        >>> assert response.status == HTTPStatus.OK
        >>> assert await response.json() == {"message": "Success"}

        Context manager pattern:
        >>> response = build_mocked_aiohttp_response(status=HTTPStatus.OK, json={"message": "Success"})
        >>> # async with session.get(url) as response:
        >>> assert response.status == HTTPStatus.OK

        Binary content:
        >>> response = build_mocked_aiohttp_response(status=HTTPStatus.OK, read=b"binary data")
        >>> data = await response.read()
        >>> assert data == b"binary data"
    """
    mock_response = MagicMock(spec=aiohttp.ClientResponse)
    mock_response.status = status
    mock_response.reason = reason or status.phrase

    # Configure ok property based on status code (True if status < 400)
    mock_response.ok = status < HTTPStatus.BAD_REQUEST

    # Configure async context manager support for `async with session.get() as response:` pattern
    mock_response.__aenter__ = AsyncMock(return_value=mock_response)
    mock_response.__aexit__ = AsyncMock(return_value=None)

    # Configure raise_for_status based on status code
    if status >= HTTPStatus.BAD_REQUEST:
        message = error_message or f"HTTP {status.value} {status.phrase}"
        mock_response.raise_for_status = MagicMock(
            side_effect=aiohttp.ClientResponseError(
                status=status,
                request_info=MagicMock(spec=aiohttp.RequestInfo),
                history=(),
                message=message,
            )
        )
    else:
        mock_response.raise_for_status = MagicMock()

    # Configure response body methods
    if json is not None:
        mock_response.json = AsyncMock(return_value=json)
    if text is not None:
        mock_response.text = AsyncMock(return_value=text)

    # Configure binary content reading - response.read() is the proper aiohttp method
    # response.content in aiohttp is a StreamReader property, not a callable
    if read is not None:
        mock_response.read = AsyncMock(return_value=read)
        # Also set up content as a mock StreamReader with a read() method for compatibility
        mock_content = MagicMock()
        mock_content.read = AsyncMock(return_value=read)
        mock_response.content = mock_content

    # Configure headers - as a dict property for context manager pattern
    if headers is not None:
        mock_response.headers = headers

    # Configure cookies
    if cookies is not None:
        cookie_obj = SimpleCookie()
        for name, value in cookies.items():
            cookie_obj[name] = value
        mock_response.cookies = cookie_obj

    # Configure optional response metadata
    if url is not None:
        mock_response.url = url
    if method is not None:
        mock_response.method = method

    return mock_response


# Type alias for parametric response types
ParametricResponse = aiohttp.ClientResponse | None | Callable[..., aiohttp.ClientResponse | None]


def build_mocked_aiohttp_resource(  # noqa: PLR0913
    get: ParametricResponse = None,
    post: ParametricResponse = None,
    put: ParametricResponse = None,
    patch: ParametricResponse = None,
    delete: ParametricResponse = None,
    head: ParametricResponse = None,
    options: ParametricResponse = None,
) -> AioHttpClientResource:
    """Build the mocked Aiohttp resource.

    Creates a mock AioHttpClientResource object with configurable HTTP method responses.
    The mock resource provides an async context manager via `acquire_client_session()` that
    returns a mock ClientSession with the specified method responses configured.

    The resource supports the context manager pattern used by services:
    ```python
    async with resource.acquire_client_session() as session:
        async with session.get(url="...") as response:
            data = await response.json()
    ```

    Supports parametric responses for multiple calls:
    - Single response: Returns the same response for all calls
    - Callable: Calls the function for each request to generate dynamic responses

    Args:
        get: Mock response to return for GET requests. Can be:
            - A single ClientResponse or None
            - A callable that takes request arguments and returns a response
        post: Mock response to return for POST requests. Same format as get.
        put: Mock response to return for PUT requests. Same format as get.
        patch: Mock response to return for PATCH requests. Same format as get.
        delete: Mock response to return for DELETE requests. Same format as get.
        head: Mock response to return for HEAD requests. Same format as get.
        options: Mock response to return for OPTIONS requests. Same format as get.

    Returns:
        AioHttpClientResource: A mocked AioHttpClientResource object configured with
            the specified HTTP method responses.

    Example:
        Context manager pattern (used by services):
        >>> get_response = build_mocked_aiohttp_response(status=HTTPStatus.OK, json={"message": "Success"})
        >>> resource = build_mocked_aiohttp_resource(get=get_response)
        >>> async with resource.acquire_client_session() as session:
        ...     async with session.get(url="https://example.com") as response:
        ...         assert response.status == HTTPStatus.OK
        ...         data = await response.json()

        Dynamic response (callable):
        >>> def get_response(url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        ...     if "page=1" in url:
        ...         return build_mocked_aiohttp_response(status=HTTPStatus.OK, json={"page": 1})
        ...     return build_mocked_aiohttp_response(status=HTTPStatus.NOT_FOUND)
        >>> resource = build_mocked_aiohttp_resource(get=get_response)
        >>> async with resource.acquire_client_session() as session:
        ...     async with session.get(url="https://example.com?page=1") as response:
        ...         assert response.status == HTTPStatus.OK
    """
    mock_resource = MagicMock(spec=AioHttpClientResource)
    session = MagicMock()

    # Configure HTTP methods dynamically to reduce repetition
    # HTTP methods return context managers (the response), not awaitables
    # This supports the pattern: `async with session.get(url) as response:`
    http_methods: dict[str, ParametricResponse] = {
        "get": get,
        "post": post,
        "put": put,
        "patch": patch,
        "delete": delete,
        "head": head,
        "options": options,
    }
    for method_name, response in http_methods.items():
        if isinstance(response, types.FunctionType):
            # Callable response: call the function to generate responses dynamically
            mock_method = MagicMock(side_effect=response)
        else:
            # Static response: return the same response for all calls
            mock_method = MagicMock(return_value=response)
        setattr(session, method_name, mock_method)

    # Configure async context manager for acquire_client_session
    # acquire_client_session() is an @asynccontextmanager, so it returns an async context manager
    # We need to create a context manager object that yields the session
    context_manager = MagicMock()
    context_manager.__aenter__ = AsyncMock(return_value=session)
    context_manager.__aexit__ = AsyncMock(return_value=None)
    # acquire_client_session is not async itself, it returns an async context manager
    mock_resource.acquire_client_session = MagicMock(return_value=context_manager)

    return mock_resource



"""Provides utilities to handle Redis Configuration."""

from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, ValidationError

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError
from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import UnableToReadYamlFileError, YamlFileReader


class RedisCredentialsConfigError(FastAPIFactoryUtilitiesError):
    """Redis credentials config error."""


class RedisCredentialsConfig(BaseModel):
    """Redis credentials config."""

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")
    url: str


def build_redis_credentials_config(application: "ApplicationAbstractProtocol") -> RedisCredentialsConfig:
    """Build the Redis credentials configuration.

    Args:
        application: The application.

    Returns:
        RedisCredentialsConfig: The Redis credentials configuration.

    Raises:
        TaskiqPluginConfigError: for any error occurred while building the Redis credentials configuration.
    """
    if application.PACKAGE_NAME == "":
        raise RedisCredentialsConfigError("The package name must be set in the concrete application class.")

    try:
        yaml_file_content: dict[str, Any] = YamlFileReader(
            file_path=get_path_file_in_package(
                filename="application.yaml",
                package=application.PACKAGE_NAME,
            ),
            yaml_base_key="redis",
            use_environment_injection=True,
        ).read()
    except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
        raise RedisCredentialsConfigError("Unable to read the application configuration file.") from exception

    try:
        config: RedisCredentialsConfig = RedisCredentialsConfig.model_validate(yaml_file_content)
    except ValidationError as exception:
        raise RedisCredentialsConfigError("Unable to create the application configuration model.") from exception
    return config



"""Package for books entities."""

from .entities import BookEntity
from .enums import BookType
from .types import BookEntityId, BookName

__all__: list[str] = ["BookEntity", "BookEntityId", "BookName", "BookType"]



"""Provide Kratos Session and Identity classes."""

from http import HTTPStatus
from typing import Generic, TypeVar

from fastapi import HTTPException, Request
from pydantic import BaseModel

from fastapi_factory_utilities.core.security.abstracts import AuthenticationAbstract
from fastapi_factory_utilities.core.services.kratos import (
    KratosGenericWhoamiService,
    KratosOperationError,
    KratosSessionInvalidError,
)

GenericKratosSessionObject = TypeVar("GenericKratosSessionObject", bound=BaseModel)


class KratosSessionAuthenticationService(AuthenticationAbstract, Generic[GenericKratosSessionObject]):
    """Kratos Session class."""

    DEFAULT_COOKIE_NAME: str = "ory_kratos_session"

    def __init__(
        self,
        kratos_service: KratosGenericWhoamiService[GenericKratosSessionObject],
        cookie_name: str = DEFAULT_COOKIE_NAME,
        raise_exception: bool = True,
    ) -> None:
        """Initialize the KratosSessionAuthentication class.

        Args:
            kratos_service (KratosGenericWhoamiService[GenericKratosSessionObject]): Kratos service object.
            cookie_name (str): Name of the cookie to extract the session.
            raise_exception (bool): Whether to raise an exception or return None.
        """
        self._kratos_service: KratosGenericWhoamiService[GenericKratosSessionObject] = kratos_service
        self._cookie_name: str = cookie_name
        self._session: GenericKratosSessionObject
        super().__init__(raise_exception=raise_exception)

    def _extract_cookie(self, request: Request) -> str | None:
        """Extract the cookie from the request.

        Args:
            request (Request): FastAPI request object.

        Returns:
            str | None: Cookie value or None if not found.

        Raises:
            HTTPException: If the cookie is missing.
        """
        return request.cookies.get(self._cookie_name, None)

    @property
    def session(self) -> GenericKratosSessionObject:
        """Get the Kratos session.

        Returns:
            GenericKratosSessionObject: Kratos session object.
        """
        return self._session

    async def authenticate(self, request: Request) -> None:
        """Extract the Kratos session from the request.

        Args:
            request (Request): FastAPI request object.

        Returns:
            None: If the authentication is successful or not raise_exception is False.

        Raises:
            HTTPException: If the session is invalid and raise_exception is True.
        """
        cookie: str | None = self._extract_cookie(request)
        if not cookie:
            return self.raise_exception(
                HTTPException(
                    status_code=HTTPStatus.UNAUTHORIZED,
                    detail="Missing Credentials",
                )
            )

        try:
            self._session = await self._kratos_service.whoami(cookie_value=cookie)
        except KratosSessionInvalidError:
            return self.raise_exception(
                HTTPException(
                    status_code=HTTPStatus.UNAUTHORIZED,
                    detail="Invalid Credentials",
                )
            )
        except KratosOperationError:
            return self.raise_exception(
                HTTPException(
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    detail="Internal Server Error",
                )
            )

        return



"""Provides the request objects for the books API."""

from pydantic import BaseModel, Field

from fastapi_factory_utilities.example.entities.books import BookName, BookType


class BookCreateRequest(BaseModel):
    """Book creation request model."""

    title: BookName = Field(title="Title of the book", description="The title of the book to create")
    book_type: BookType = Field(title="Type of book", description="The genre/type of the book")


class BookUpdateRequest(BaseModel):
    """Book update request model."""

    title: BookName = Field(title="Title of the book", description="The updated title of the book")
    book_type: BookType = Field(title="Type of book", description="The updated genre/type of the book")



"""Provides the exceptions for the Aiohttp plugin."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class AioHttpClientError(FastAPIFactoryUtilitiesError):
    """Exception for the Aiohttp client."""


class UnableToReadHttpDependencyConfigError(AioHttpClientError):
    """Exception for the unable to read the HTTP dependency config."""


class AioHttpClientResourceNotFoundError(AioHttpClientError):
    """Exception for the Aiohttp resource not found in the application state."""



"""Provides the JWK stores."""

from abc import ABC, abstractmethod
from asyncio import Lock

from fastapi import Request
from fastapi.datastructures import State
from jwt import PyJWK, PyJWKSet

from fastapi_factory_utilities.core.security.types import OAuth2Issuer
from fastapi_factory_utilities.core.services.hydra import (
    HydraIntrospectGenericService,
    HydraOperationError,
    HydraTokenIntrospectObject,
)

from .exceptions import HydraJWKSStoreError


class JWKStoreAbstract(ABC):
    """JWK store abstract class."""

    def __init__(self) -> None:
        """Initialize the JWK store."""
        self._issuer_by_kid: dict[str, OAuth2Issuer] = {}
        self._jwk_by_kid: dict[str, PyJWK] = {}

    @abstractmethod
    async def get_jwk(self, kid: str) -> PyJWK:
        """Get the JWK from the store."""
        raise NotImplementedError()

    @abstractmethod
    async def get_issuer_by_kid(self, kid: str) -> OAuth2Issuer:
        """Get the issuer by kid from the store."""
        raise NotImplementedError()

    @abstractmethod
    async def get_jwks(self, issuer: OAuth2Issuer) -> PyJWKSet:
        """Get the JWKS from the store."""
        raise NotImplementedError()

    @abstractmethod
    async def add_jwk(self, issuer: OAuth2Issuer, jwk: PyJWK) -> None:
        """Add a JWK to the store."""
        raise NotImplementedError()


class JWKStoreMemory(JWKStoreAbstract):
    """JWK store in memory. Concurrent safe."""

    def __init__(self) -> None:
        """Initialize the JWK store in memory."""
        super().__init__()
        self._lock: Lock = Lock()

    async def get_jwk(self, kid: str) -> PyJWK:
        """Get the JWK from the store."""
        async with self._lock:
            return self._jwk_by_kid[kid]

    async def get_issuer_by_kid(self, kid: str) -> OAuth2Issuer:
        """Get the issuer by kid from the store."""
        async with self._lock:
            return self._issuer_by_kid[kid]

    async def get_jwks(self, issuer: OAuth2Issuer) -> PyJWKSet:
        """Get the JWKS from the store."""
        async with self._lock:
            jwks: list[PyJWK] = []
            for kid, jwk in self._jwk_by_kid.items():
                if self._issuer_by_kid[kid] == issuer:
                    jwks.append(jwk)
            return PyJWKSet.from_dict(
                {
                    "keys": [jwk._jwk_data for jwk in jwks]  # pylint: disable=protected-access # pyright: ignore[reportPrivateUsage]
                }
            )

    async def add_jwk(self, issuer: OAuth2Issuer, jwk: PyJWK) -> None:
        """Add a JWK to the store."""
        async with self._lock:
            if jwk.key_id is None:
                raise ValueError("JWK key ID is required")
            self._jwk_by_kid[jwk.key_id] = jwk
            self._issuer_by_kid[jwk.key_id] = issuer


async def configure_jwks_in_memory_store_from_hydra_introspect_services(
    introspect_service_list: list[HydraIntrospectGenericService[HydraTokenIntrospectObject]],
) -> JWKStoreMemory:
    """Configure the JWKS in memory store from the Hydra introspect services."""
    jwk_store = JWKStoreMemory()
    try:
        for introspect_service in introspect_service_list:
            jwks: list[PyJWK] = await introspect_service.get_wellknown_jwks()
            for key in jwks:
                assert key.key_id is not None
                await jwk_store.add_jwk(introspect_service.get_issuer(), key)
    except (HydraOperationError, RuntimeError) as e:
        raise HydraJWKSStoreError("Failed to get the JWKS from the introspect services") from e
    return jwk_store


class DependsHydraJWKStoreMemory:
    """Dependency for the Hydra JWKS store in memory."""

    DEPENDENCY_KEY: str = "hydra_jwks_store_memory"

    @classmethod
    def export_from_state(cls, state: State) -> JWKStoreMemory:
        """Export the Hydra JWKS store in memory from the state."""
        jwk_store: JWKStoreMemory | None = getattr(state, cls.DEPENDENCY_KEY, None)
        if jwk_store is None:
            raise HydraJWKSStoreError("Hydra JWKS store in memory not found in the state")
        return jwk_store

    @classmethod
    def import_to_state(cls, state: State, jwk_store: JWKStoreMemory) -> None:
        """Import the Hydra JWKS store in memory to the state."""
        setattr(state, cls.DEPENDENCY_KEY, jwk_store)

    def __call__(self, request: Request) -> JWKStoreMemory:
        """Dependency for the Hydra JWKS store in memory."""
        return self.export_from_state(state=request.app.state)
